from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.lqp.utils import UniqueNames
from relationalai.early_access.metamodel import ir

rel_to_lqp = {
    "+": "rel_primitive_add",
    "-": "rel_primitive_subtract",
    "*": "rel_primitive_multiply",
    "/": "rel_primitive_divide",
    "=": "rel_primitive_eq",
    "!=": "rel_primitive_neq",
    "<=": "rel_primitive_lt_eq",
    ">=": "rel_primitive_gt_eq",
    ">": "rel_primitive_gt",
    "<": "rel_primitive_lt",
    "abs": "rel_primitive_abs",
    "construct_date": "rel_primitive_construct_date",
    "construct_datetime": "rel_primitive_construct_datetime",
    "starts_with": "rel_primitive_starts_with",
    "ends_with": "rel_primitive_ends_with",
    "contains": "rel_primitive_contains",
    "num_chars": "rel_primitive_num_chars",
    "substring": "rel_primitive_substring",
    "like_match": "rel_primitive_like_match",
    "concat": "rel_primitive_concat",
    "replace": "rel_primitive_replace",
    "date_year": "rel_primitive_date_year",
    "date_month": "rel_primitive_date_month",
    "date_day": "rel_primitive_date_day",
}

def relname_to_lqp_name(name: str) -> str:
    # TODO: do these proprly
    if name in rel_to_lqp:
        return rel_to_lqp[name]
    else:
        raise NotImplementedError(f"missing primitive case: {name}")

def lqp_sum_op(names: UniqueNames) -> lqp.Abstraction:
    # TODO: make sure gensym'd properly
    x = lqp.Var(names.get_name("x"))
    y = lqp.Var(names.get_name("y"))
    z = lqp.Var(names.get_name("z"))
    # TODO: Extract a more refined type from the op relation
    INT = lqp.PrimitiveType.INT
    ts = [(x, INT), (y, INT), (z, INT)]

    body = lqp.Primitive("rel_primitive_add", [x, y, z])
    return lqp.Abstraction(ts, body)

def lqp_max_op(names: UniqueNames) -> lqp.Abstraction:
    # TODO: make sure gensym'd properly
    x = lqp.Var(names.get_name("x"))
    y = lqp.Var(names.get_name("y"))
    z = lqp.Var(names.get_name("z"))
    # TODO: Extract a more refined type from the op relation
    INT = lqp.PrimitiveType.INT
    ts = [(x, INT), (y, INT), (z, INT)]

    body = lqp.Primitive("rel_primitive_max", [x, y, z])
    return lqp.Abstraction(ts, body)

def lqp_min_op(names: UniqueNames) -> lqp.Abstraction:
    # TODO: make sure gensym'd properly
    x = lqp.Var(names.get_name("x"))
    y = lqp.Var(names.get_name("y"))
    z = lqp.Var(names.get_name("z"))
    # TODO: Extract a more refined type from the op relation
    INT = lqp.PrimitiveType.INT
    ts = [(x, INT), (y, INT), (z, INT)]

    body = lqp.Primitive("rel_primitive_min", [x, y, z])
    return lqp.Abstraction(ts, body)

def lqp_operator(names: UniqueNames, op: ir.Relation) -> lqp.Abstraction:
    if op.name == "sum":
        return lqp_sum_op(names)
    elif op.name == "count":
        return lqp_sum_op(names)
    elif op.name == "max":
        return lqp_max_op(names)
    elif op.name == "min":
        return lqp_min_op(names)
    else:
        raise NotImplementedError(f"Unsupported aggregation: {op.name}")



"""
Helpers to analyze the metamodel IR in ways that are specific to Rel generation.
"""
from __future__ import annotations

from typing import cast, Optional
from relationalai.early_access.metamodel import ir, helpers, factory as f
from relationalai.early_access.metamodel.util import OrderedSet
from relationalai.early_access.rel.dependency import AnalysisContext


def outer_join_prefix(nullable_logicals: OrderedSet[ir.Logical], groups: dict[str, OrderedSet[ir.Task]]) -> list[ir.Var]:
    """ Check if it is possible to use an outer join on the logical that has these nullable
    logicals and these groups of tasks. This function returns a list of prefix variables
    for the outer join.

    There are several requirements to use the outer join. If some is not met, the
    function returns an empty list
    """
    # all nested logicals must be nullable
    if len(nullable_logicals) != len(groups["logicals"]):
        return []

    # at this point, all nested logicals are nullable, but they should only have a single hoisted variable
    for logical in nullable_logicals:
        if len(logical.hoisted) != 1:
            return []

    # outer joins only work on outputs
    if "output" not in groups or len(groups["output"]) != 1:
        return []
    output = cast(ir.Output, groups["output"].some())

    # the output variables must be a prefix + nullable variables; the length of the
    # prefix is the number of extra aliases in the output
    prefix_length = len(output.aliases) - len(nullable_logicals)
    if prefix_length < 1:
        return []

    # get the hoisted vars from the nullable logicals, a bit messy
    logical_hoisted_vars = OrderedSet.from_iterable([cast(ir.Default, logical.hoisted[0]).var for logical in nullable_logicals])
    prefix=[]
    i = 0
    for _, var in output.aliases:
        if i < prefix_length:
            # the first prefix_length aliases are the prefix variables
            prefix.append(var)
            i += 1
        else:
            # the remaining output variables should be exposed one by one by the logicals
            if var not in logical_hoisted_vars:
                return []
            logical_hoisted_vars.remove(var)
    # basically an assertion as this should be empty if we got here
    if logical_hoisted_vars:
        return []

    # all nullable logicals must join with all the prefix variables
    for logical in nullable_logicals:
        vars = helpers.collect_implicit_vars(logical)
        for prefix_var in prefix:
            if prefix_var not in vars:
                return []

    return prefix


def extract(task: ir.Task, body: OrderedSet[ir.Task], exposed_vars: list[ir.Var], ctx: AnalysisContext, prefix: Optional[str]=None) -> ir.Relation:
    """
    Extract into this Analysiscontext a new top level Logical that contains this body plus a
    derive task into a new temporary relation, which is also registered with the ctx.
    The exposed_vars determine the arguments of this temporary relation. The prefix
    can be used to customize the name of the relation, which defaults to the task kind.

    Return the temporary relation created for the extraction.
    """

    # TODO: review variable rewrites, i.e. when we extract a new logical, we should remap variables
    p = prefix if prefix else task.kind

    # new relation to derive the aggregation into
    connection = f.relation(f"_{p}_{task.id}", [f.field(v.name, v.type) for v in exposed_vars])
    ctx.relations.append(connection)

    # add derivation to the extracted body
    body.add(f.derive(connection, exposed_vars))

    # extract the body
    ctx.top_level.append(ir.Logical(task.engine, tuple(), tuple(body)))

    # return a reference to the connection relation
    return connection



from relationalai.early_access.lqp.validators import assert_valid_input
from relationalai.early_access.lqp.utils import UniqueNames
from relationalai.early_access.metamodel import ir, builtins as rel_builtins, helpers
from relationalai.early_access.metamodel.types import is_any
from relationalai.early_access.metamodel.visitor import collect_by_type
from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.lqp.hash_utils import lqp_hash
from relationalai.early_access.lqp.primitives import relname_to_lqp_name, lqp_operator
from relationalai.early_access.lqp.types import meta_type_to_lqp, type_from_constant
from relationalai.early_access.lqp.constructors import mk_and, mk_exists, mk_or, mk_abstraction
import datetime as dt

from typing import Tuple, cast, Union

class TranslationCtx:
    def __init__(self, model):
        # TODO: comment these fields
        # TODO: should we have a pass to rename variables instead of this?
        self.unique_names = UniqueNames()
        self.id_to_orig_name = {}
        self.output_ids = []

""" Main access point. Converts the model IR to an LQP program. """
def to_lqp(model: ir.Model) -> lqp.LqpProgram:
    assert_valid_input(model)
    ctx = TranslationCtx(model)
    program = _translate_to_program(ctx, model)
    return program

def _translate_to_program(ctx: TranslationCtx, model: ir.Model) -> lqp.LqpProgram:
    decls: list[lqp.Declaration] = []
    outputs: list[Tuple[str, lqp.RelationId]] = []

    # LQP only accepts logical tasks
    # These are asserted at init time
    root = cast(ir.Logical, model.root)

    seen_rids = set()
    for subtask in root.body:
        assert isinstance(subtask, ir.Logical)
        decl = _translate_to_decl(ctx, subtask)

        if decl is None:
            continue

        assert isinstance(decl, lqp.Def), "we dont do loops yet m8"
        decls.append(decl)

        rid = decl.name
        assert rid not in seen_rids, f"duplicate relation id: {rid}"
        seen_rids.add(rid)

    for output_id in ctx.output_ids:
        assert isinstance(output_id, lqp.RelationId)
        outputs.append(("output", output_id))

    debug_info = lqp.DebugInfo(ctx.id_to_orig_name)

    return lqp.LqpProgram(decls, outputs, debug_info)

def _make_def(ctx: TranslationCtx, name: str, projection: list[Tuple[lqp.Var, lqp.RelType]], conjs: list[lqp.Formula], is_output: bool = False) -> lqp.Def:
    abstraction = mk_abstraction(projection, mk_and(conjs))
    rel_id = name_to_id(name)
    ctx.id_to_orig_name[rel_id] = name

    if is_output:
        ctx.output_ids.append(rel_id)

    # TODO: is this correct? might need attrs tooo?
    return lqp.Def(rel_id, abstraction, [])

def _effect_bindings(effect: Union[ir.Output, ir.Update]) -> list[ir.Value]:
    if isinstance(effect, ir.Output):
        # TODO: we dont yet handle aliases, so we ignore v[0]
        return [v[1] for v in effect.aliases]
    else:
        return list(effect.args)

def _translate_to_decl(ctx: TranslationCtx, rule: ir.Logical) -> Union[lqp.Declaration,None]:
    effects = collect_by_type((ir.Output, ir.Update), rule)
    aggregates = collect_by_type(ir.Aggregate, rule)

    # TODO: should this ever actually come in as input?
    if len(effects) == 0:
        return None
    assert len(effects) == 1, f"should only have exactly one effect, got {len(effects)}"
    effect = effects[0]

    conjuncts = []
    for task in rule.body:
        if isinstance(task, (ir.Output, ir.Update)):
            continue
        conjuncts.append(_translate_to_formula(ctx, task))

    # Aggregates reduce over the body
    if len(aggregates) > 0:
        aggr_body = mk_and(conjuncts)
        conjuncts = []
        for aggr in aggregates:
            conjuncts.append(_translate_aggregate(ctx, aggr, aggr_body))

    # Handle the bindings
    bindings = _effect_bindings(effect)
    projection, eqs = translate_bindings(ctx, bindings)
    conjuncts.extend(eqs)

    is_output = isinstance(effect, ir.Output)
    def_name = "output" if is_output else effect.relation.name
    return _make_def(ctx, def_name, projection, conjuncts, is_output)

def _translate_aggregate(ctx: TranslationCtx, aggr: ir.Aggregate, body: lqp.Formula) -> lqp.Reduce:
    # TODO: handle this properly
    aggr_name = aggr.aggregation.name
    supported_aggrs = ("sum", "count", "min", "max")
    assert aggr_name in supported_aggrs, f"only support {supported_aggrs} for now, not {aggr.aggregation.name}"

    # TODO: This is not right, we need to handle input args and output args properly
    # Last one is output arg, the rest are input args
    input_args = [_translate_term(ctx, arg) for arg in aggr.args[:-1]]
    output_var, _ = _translate_term(ctx, aggr.args[-1])

    projected_args = [_translate_term(ctx, var) for var in aggr.projection]
    abstr_args = []
    abstr_args.extend(projected_args)
    abstr_args.extend(input_args)
    if aggr_name == "count":
        # Count sums up "1"
        one_var, typ, eq = binding_to_lqp_var(ctx, 1)
        assert eq is not None
        body = mk_and([body, eq])
        abstr_args.append([one_var, typ])

    # Group-bys do not need to be handled at all, since they are introduced outside already
    reduce = lqp.Reduce(
        lqp_operator(ctx.unique_names, aggr.aggregation),
        mk_abstraction(abstr_args, body),
        [output_var],
    )
    return reduce

def _translate_to_formula(ctx: TranslationCtx, task: ir.Task) -> lqp.Formula:
    if isinstance(task, ir.Logical):
        conjuncts = [_translate_to_formula(ctx, child) for child in task.body]
        return mk_and(conjuncts)
    elif isinstance(task, ir.Lookup):
        return _translate_to_atom(ctx, task)
    elif isinstance(task, ir.Not):
        return lqp.Not(_translate_to_formula(ctx, task.task))
    elif isinstance(task, ir.Exists):
        lqp_vars, conjuncts = translate_bindings(ctx, list(task.vars))
        conjuncts.append(_translate_to_formula(ctx, task.task))
        return mk_exists(lqp_vars, mk_and(conjuncts))
    elif isinstance(task, ir.Construct):
        assert len(task.values) >= 1, "construct should have at least one value"
        # TODO: what does the first value do
        terms = [_translate_term(ctx, arg) for arg in task.values[1:]]
        terms.append(_translate_term(ctx, task.id_var))
        return lqp.Primitive(
            "rel_primitive_hash_tuple_uint128",
            [v for [v, _] in terms],
        )
    elif isinstance(task, ir.Union):
        # TODO: handle hoisted vars if needed
        assert len(task.hoisted) == 0, "hoisted updates not supported yet, because idk what it means"
        disjs = [_translate_to_formula(ctx, child) for child in task.tasks]
        return mk_or(disjs)
    elif isinstance(task, ir.Aggregate):
        # Nothing to do here, handled in _translate_to_decl
        return mk_and([])
    else:
        raise NotImplementedError(f"Unknown task type (formula): {type(task)}")

def _translate_term(ctx: TranslationCtx, value: ir.Value) -> Tuple[lqp.Term, lqp.RelType]:
    if isinstance(value, ir.Var):
        name = ctx.unique_names.get_name_by_id(value.id, value.name)
        assert not is_any(value.type), f"unexpected type for var `{value}`: {value.type}"
        t = meta_type_to_lqp(value.type)
        return lqp.Var(name), t
    elif isinstance(value, ir.Literal):
        return _translate_term(ctx, value.value)
    else:
        assert isinstance(value, (str, int, dt.date)), f"unexpected value type: {type(value)}"
        return lqp.Constant(value), type_from_constant(value)

def _translate_to_atom(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    # TODO: want signature not name
    rel_name = task.relation.name
    terms = []
    sig_types = []
    for arg in task.args:
        if isinstance(arg, ir.PyValue):
            assert isinstance(arg, lqp.PrimitiveValue)
            term = lqp.Constant(arg)
            terms.append(term)
            t = type_from_constant(arg)
            sig_types.append(t)
            continue
        elif isinstance(arg, ir.Literal):
            term = lqp.Constant(arg.value)
            terms.append(term)
            t = type_from_constant(arg.value)
            sig_types.append(t)
            continue
        assert isinstance(arg, ir.Var), f"expected var, got {type(arg)}: {arg}"
        term, t = _translate_term(ctx, arg)
        terms.append(term)
        sig_types.append(t)

    # TODO: wrong
    if rel_builtins.is_builtin(task.relation):
        lqp_name = relname_to_lqp_name(task.relation.name)
        return lqp.Primitive(lqp_name, terms)

    if helpers.is_external(task.relation):
        return lqp.RelAtom(
            task.relation.name,
            terms,
        )

    rid = get_relation_id(ctx, rel_name, sig_types)
    return lqp.Atom(rid, terms)

def get_relation_id(ctx: TranslationCtx, name: str, types: list[lqp.PrimitiveType]) -> lqp.RelationId:
    relation_id = name_to_id(name)
    ctx.id_to_orig_name[relation_id] = name
    return relation_id

# TODO: should this take types too?
# TODO: should we use unique numbers instead of hashes?
def name_to_id(name: str) -> lqp.RelationId:
    return lqp.RelationId(lqp_hash(name))

def translate_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.RelType]], list[lqp.Formula]]:
    lqp_vars = []
    conjuncts = []
    for binding in bindings:
        lqp_var, typ, eq = binding_to_lqp_var(ctx, binding)
        lqp_vars.append((lqp_var, typ))
        if eq is not None:
            conjuncts.append(eq)

    return lqp_vars, conjuncts

def binding_to_lqp_var(ctx: TranslationCtx, binding: ir.Value) -> Tuple[lqp.Var, lqp.RelType, lqp.Union[None, lqp.Formula]]:
    if isinstance(binding, ir.Var):
        var, typ = _translate_term(ctx, binding)
        assert isinstance(var, lqp.Var)
        return var, typ, None
    else:
        # Constant in this case
        assert isinstance(binding, lqp.PrimitiveValue)
        value, typ = _translate_term(ctx, binding)

        # TODO: gensym
        var_name = ctx.unique_names.get_name_by_id(1, "cvar")
        var = lqp.Var(var_name)
        eq = lqp.Primitive("rel_primitive_eq", [var, value])
        return var, typ, eq



from relationalai.early_access.metamodel.compiler import Pass
from relationalai.early_access.metamodel import ir, builtins as rel_builtins, factory as f, typer, visitor
from typing import cast, Tuple, Union
from relationalai.early_access.metamodel.util import NameCache
from relationalai.early_access.metamodel import types

from relationalai.early_access.metamodel.rewrite import Splinter, RewriteListTypes, GarbageCollectNodes

# TODO: Move this into metamodel.rewrite
from relationalai.early_access.rel.rewrite import Flatten, QuantifyVars, ExtractCommon, CDC

import datetime

def lqp_passes() -> list[Pass]:
    # TODO: are these all needed from the Rel emitter?
    # TODO: should there be a pass to remove aliases from output?
    return [
        RewriteListTypes(),
        GarbageCollectNodes(),
        CDC(),
        typer.Checker(),
        typer.Typer(),
        GarbageCollectNodes(),
        ExtractCommon(),
        Flatten(),
        QuantifyVars(),
        Splinter(), # Adds missing existentials + splits multi-headed rules into single rules
        NormalizeAggregates(),
        HoistAggregates(),
        UnifyDefinitions(),
        EliminateValueTypeConstants(),
    ]

# TODO: assert theres no nested nested nested aggregates
# TODO: assert theres only one aggregate per rule
# TODO: is this correct?
# Move aggregates to the same level as updates (i.e. top-level logical)
class HoistAggregates(Pass):
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        root = cast(ir.Logical, model.root)
        new_subtasks = []
        for subtask in root.body:
            subtask = cast(ir.Logical, subtask)
            new_subsubtasks = []
            for subsubtask in subtask.body:
                if not isinstance(subsubtask, ir.Logical):
                    new_subsubtasks.append(subsubtask)
                    continue

                subsubtask = cast(ir.Logical, subsubtask)
                if self.has_aggregate(subsubtask):
                    new_subsubtask, aggrs = self._hoist_aggregates(subsubtask)
                    new_subsubtasks.append(new_subsubtask)
                    new_subsubtasks.extend(aggrs)
                else:
                    new_subsubtasks.append(subsubtask)

            new_subtask = ir.Logical(
                    subtask.engine,
                    subtask.hoisted,
                    tuple(new_subsubtasks),
                )
            new_subtasks.append(new_subtask)

        new_root = ir.Logical(root.engine, root.hoisted, tuple(new_subtasks))
        model = ir.Model(
            model.engines,
            model.relations,
            model.types,
            new_root,
        )
        return model

    def has_aggregate(self, task: ir.Task) -> bool:
        assert isinstance(task, ir.Logical)
        for subtask in task.body:
            if isinstance(subtask, ir.Aggregate):
                return True
        return False

    def _hoist_aggregates(self, task: ir.Task) -> Tuple[ir.Task, list[ir.Aggregate]]:
        assert isinstance(task, ir.Logical)
        aggrs = []
        new_subtasks = []
        for subtask in task.body:
            if isinstance(subtask, ir.Aggregate):
                aggrs.append(subtask)
            else:
                new_subtasks.append(subtask)
        assert len(aggrs) > 0, "should have found an aggregate"
        return (ir.Logical(
            task.engine,
            task.hoisted,
            tuple(new_subtasks),
        ), aggrs)

# Rewrites aggregates when necessary e.g. avg -> sum/count
class NormalizeAggregates(Pass):
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        self.var_name_cache: NameCache = NameCache()
        root = cast(ir.Logical, model.root)
        new_subtasks = []
        for subtask in root.body:
            subtask = cast(ir.Logical, subtask)
            if self.has_aggregate(subtask):
                new_subtask = self._rewrite_subtask(subtask)
                new_subtasks.append(new_subtask)
            else:
                new_subtasks.append(subtask)
        new_root = ir.Logical(root.engine, root.hoisted, tuple(new_subtasks))
        model = ir.Model(
            model.engines,
            model.relations,
            model.types,
            new_root,
        )
        return model

    def _rewrite_subtask(self, task: ir.Task) -> ir.Task:
        assert isinstance(task, ir.Logical)
        new_subtasks = []
        for subtask in task.body:
            if isinstance(subtask, ir.Aggregate):
                assert isinstance(subtask, ir.Aggregate)
                if subtask.aggregation.name == "sum":
                    new_subtasks.append(subtask)
                elif subtask.aggregation.name == "count":
                    new_subtasks.append(subtask)
                elif subtask.aggregation.name == "avg":
                    avg_task = self._make_avg(subtask)
                    new_subtasks.append(avg_task)
                elif subtask.aggregation.name == "max":
                    new_subtasks.append(subtask)
                elif subtask.aggregation.name == "min":
                    new_subtasks.append(subtask)
                else:
                    raise NotImplementedError(f"Unsupported aggregate: {subtask.aggregation.name}")
            else:
                new_subtasks.append(subtask)

        return ir.Logical(
            task.engine,
            task.hoisted,
            tuple(new_subtasks),
        )

    def _make_avg(self, task: ir.Aggregate) -> ir.Exists:
        assert isinstance(task, ir.Aggregate)
        assert task.aggregation.name == "avg"

        # TODO: confirm input args
        assert len(task.args) == 2
        input_arg = task.args[0]
        output_arg = task.args[1]
        assert isinstance(input_arg, ir.Var), f"expected var, got {type(input_arg)}: {input_arg}"
        assert isinstance(output_arg, ir.Var), f"expected var, got {type(output_arg)}: {output_arg}"

        new_sum_name = self.var_name_cache.get_name(input_arg.id, "sum")
        new_count_name = self.var_name_cache.get_name(input_arg.id, "count")
        new_sum = ir.Var(output_arg.type, new_sum_name)
        new_count = ir.Var(output_arg.type, new_count_name)

        sum_args = tuple([input_arg, new_sum])
        # Count doesn't have an 'over' argument
        count_args = tuple([input_arg, new_count])

        sum_task = ir.Aggregate(
            task.engine,
            rel_builtins.sum,
            task.projection,
            task.group,
            sum_args,
        )
        count_task = ir.Aggregate(
            task.engine,
            rel_builtins.count,
            task.projection,
            task.group,
            count_args,
        )

        final_result = ir.Lookup(
            task.engine,
            rel_builtins.div,
            tuple([new_sum, new_count, output_arg]),
        )

        body = ir.Logical(
            task.engine,
            tuple(),
            tuple([sum_task, count_task, final_result]),
        )

        return f.exists([new_sum, new_count], body)

    def has_aggregate(self, task: ir.Task) -> bool:
        assert isinstance(task, ir.Logical)
        for subtask in task.body:
            if isinstance(subtask, ir.Aggregate):
                return True
        return False

class UnifyDefinitions(Pass):
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        multidef_relations = self.get_multidef_relations(model)
        for relation in multidef_relations:
            model = self.rename_multidef(model, relation)
        return model

    def get_multidef_relations(self, model: ir.Model) -> set[ir.Relation]:
        seen = set()
        result = set()
        root = cast(ir.Logical, model.root)
        for task in root.body:
            task = cast(ir.Logical, task)
            for subtask in task.body:
                if isinstance(subtask, ir.Output):
                    name = "output"
                    if name in seen:
                        raise NotImplementedError("multiple outputs not supported yet")
                    seen.add(name)
                elif isinstance(subtask, ir.Update):
                    assert subtask.effect == ir.Effect.derive, "only derive updates supported yet"
                    name = subtask.relation
                    if name in seen:
                        result.add(name)
                    seen.add(name)
        return result

    def rename_multidef(self, model: ir.Model, relation: ir.Relation) -> ir.Model:
        defname = relation.name
        root = cast(ir.Logical, model.root)
        new_subtasks = []
        new_relations = []
        total_ct = 0
        for subtask in root.body:
            subtask = cast(ir.Logical, subtask)
            new_subsubtasks = []
            for subsubtask in subtask.body:
                if isinstance(subsubtask, ir.Update):
                    assert subsubtask.effect == ir.Effect.derive, "only derive updates supported yet"
                    name = subsubtask.relation
                    if name == relation:
                        total_ct += 1
                        # TODO: this needs to be unique btw (gensym)
                        new_name = f"{defname}_{total_ct}"
                        new_subsubtask = self.rename_relation(subsubtask, new_name)
                        new_subsubtasks.append(new_subsubtask)
                        new_relation = ir.Relation(
                            new_name,
                            relation.fields,
                            relation.requires,
                        )
                        new_relations.append(new_relation)
                    else:
                        new_subsubtasks.append(subsubtask)
                else:
                    new_subsubtasks.append(subsubtask)

            new_subtask = ir.Logical(
                subtask.engine,
                subtask.hoisted,
                tuple(new_subsubtasks),
            )
            new_subtasks.append(new_subtask)
        assert total_ct > 0, f"should have found at least one definition for {defname}"

        args = []
        for field in relation.fields:
            args.append(ir.Var(field.type, field.name))

        # Also add the new definition
        final_relation = ir.Relation(
            defname,
            relation.fields,
            relation.requires,
        )
        new_update = ir.Update(
            root.engine,
            final_relation,
            tuple(args),
            ir.Effect.derive,
        )

        logical_tasks = []
        lookups = []
        for new_relation in new_relations:
            new_lookup = ir.Lookup(
                root.engine,
                new_relation,
                tuple(args),
            )
            lookups.append(new_lookup)

        disj = ir.Union(
            root.engine,
            tuple(),
            tuple(lookups),
        )
        logical_tasks.append(disj)
        logical_tasks.append(new_update)
        new_logical = ir.Logical(
            root.engine,
            root.hoisted,
            tuple(logical_tasks),
        )
        new_subtasks.append(new_logical)

        new_root = ir.Logical(root.engine, root.hoisted, tuple(new_subtasks))
        model = ir.Model(
            model.engines,
            model.relations | new_relations,
            model.types,
            new_root,
        )
        return model

    def rename_relation(self, task: ir.Update, new_name: str) -> ir.Update:
        assert isinstance(task, ir.Update)
        assert task.effect == ir.Effect.derive, "only derive updates supported yet"
        old_relation = task.relation
        new_relation = ir.Relation(
            new_name,
            old_relation.fields,
            old_relation.requires,
        )
        new_task = ir.Update(
            task.engine,
            new_relation,
            task.args,
            task.effect,
        )
        return new_task

# We don't have a proto representation for these, so we rewrite them to be constructed directly
class EliminateValueTypeConstants(Pass):
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        p = RewriteBadLiterals()
        return p.walk(model)

class RewriteBadLiterals(visitor.DeprecatedPass):
    def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> Union[ir.Lookup, ir.Exists]:
        args = node.args
        vars_to_existify = []
        new_conjs = []
        new_args = []
        for arg in args:
            if isinstance(arg, datetime.datetime):
                new_var = f.var("dt_var", types.DateTime)
                new_args.append(new_var)
                vars_to_existify.append(new_var)

                year = arg.year
                month = arg.month
                day = arg.day
                hour = arg.hour
                minute = arg.minute
                second = arg.second

                lookup = f.lookup(
                        rel_builtins.construct_datetime,
                        tuple([
                            f.literal(year),
                            f.literal(month),
                            f.literal(day),
                            f.literal(hour),
                            f.literal(minute),
                            f.literal(second),
                            new_var,
                        ])
                    )
                new_conjs.append(lookup)
            elif isinstance(arg, datetime.date):
                new_var = f.var("dt_var", types.Date)
                new_args.append(new_var)
                vars_to_existify.append(new_var)

                year = arg.year
                month = arg.month
                day = arg.day

                lookup = f.lookup(
                        rel_builtins.construct_date,
                        tuple([
                            f.literal(year),
                            f.literal(month),
                            f.literal(day),
                            new_var,
                        ])
                    )
                new_conjs.append(lookup)
            else:
                new_args.append(arg)

        if len(vars_to_existify) == 0:
            return node

        new_lookup = f.lookup(
            node.relation,
            tuple(new_args),
        )
        new_conjs.append(new_lookup)

        result = f.exists(
            vars_to_existify,
            f.logical(tuple(new_conjs)),
        )

        return result



import hashlib

"""
Hashes an LQP node to a 128-bit integer.
"""
def lqp_hash(node) -> int:
    return hash_to_uint128(_lqp_hash_fn(node))

# TODO: this is NOT a good hash its just to get things working for now to get a stable id.
def _lqp_hash_fn(node) -> int:
    return int.from_bytes(hashlib.sha256(str(node).encode()).digest(), byteorder='big', signed=False)

def hash_to_uint128(h: int) -> int:
    return h % (2**128)  # Ensure it's within the 128-bit range



# pyright: reportArgumentType=false
from typing import cast, Union

from relationalai.early_access.dsl.bindings.tables import (FilteringSubtypeBinding, ConceptBinding, IdentifierBinding,
                                                           SubtypeBinding, RoleBinding, Binding, ColumnRef,
                                                           SnowflakeTable)
from relationalai.early_access.dsl.codegen.common import PotentiallyBoundRelationship, BoundExternalPreferredUC
from relationalai.early_access.dsl.core.types import AbstractValueType
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.ontologies.constraints import Unique, Mandatory
from relationalai.early_access.dsl.ontologies.relationships import Relationship
from relationalai.early_access.dsl.ontologies.roles import Role, AbstractRole
from relationalai.early_access.dsl.types.entities import EntityType
from relationalai.early_access.metamodel.util import OrderedSet


class Reasoner:
    def __init__(self, model):
        self._model = model

        # Maps for type relationships
        self._subtype_map: dict['EntityType', OrderedSet['EntityType']] = {}
        self._supertype_map: dict['EntityType', OrderedSet['EntityType']] = {}
        self._subtype_closure: dict[Type, OrderedSet[Type]] = {}
        self._supertype_closure: dict[Type, OrderedSet[Type]] = {}
        self._type_closure: dict[Type, OrderedSet[Type]] = {}

        # Bindings
        self._value_type_bindings: OrderedSet[Binding] = OrderedSet()
        self._entity_type_bindings: OrderedSet[Binding] = OrderedSet()
        self._constructor_bindings: OrderedSet[Union[IdentifierBinding, SubtypeBinding]] = OrderedSet()
        self._filtering_bindings: OrderedSet[FilteringSubtypeBinding] = OrderedSet()
        self._referent_constructor_bindings: OrderedSet[Union[IdentifierBinding, SubtypeBinding]] = OrderedSet()
        self._referent_bindings: OrderedSet[RoleBinding] = OrderedSet()
        self._subtype_bindings: OrderedSet[SubtypeBinding] = OrderedSet()
        self._subtype_binding_references: dict[SubtypeBinding, Role] = {}
        self._constructor_binding_binds_to: dict[IdentifierBinding, Role] = {}
        self._role_bound_thru: dict[Role, list[Binding]] = {}
        self._subtype_bound_thru: dict[EntityType, list[Binding]] = {}
        self._ref_binding_to_ctor_binding: dict[Binding, Binding] = {}
        self._concept_binding_to_actual: dict[Binding, Binding] = {}

        # Relationships & Ontological Constraints
        self._unary_relationships: OrderedSet['Relationship'] = OrderedSet()
        self._binary_relationships: OrderedSet['Relationship'] = OrderedSet()
        self._identifier_relationships: OrderedSet['Relationship'] = OrderedSet()
        self._entity_identifier_relationships: OrderedSet['Relationship'] = OrderedSet()
        self._entity_type_identifier: dict['EntityType', 'Relationship'] = {}
        self._bound_relationships: OrderedSet['PotentiallyBoundRelationship'] = OrderedSet()
        self._bound_external_ucs: OrderedSet['BoundExternalPreferredUC'] = OrderedSet()
        self._internal_preferred_uc: OrderedSet['Unique'] = OrderedSet()
        self._external_preferred_uc: OrderedSet['Unique'] = OrderedSet()
        self._constructor_roles: OrderedSet['AbstractRole'] = OrderedSet()
        self._internal_uc_roles: OrderedSet['Role'] = OrderedSet()
        self._mandatory_roles: OrderedSet['Role'] = OrderedSet()
        self._bound_role: dict['Role', list['Binding']] = {}
        self._ref_type_of: dict['EntityType', 'EntityType'] = {}
        self._inclusive_entity_types: OrderedSet['EntityType'] = OrderedSet()
        self._composite_entity_types: OrderedSet['EntityType'] = OrderedSet()
        self._identifiers_of_subtype: dict[EntityType, OrderedSet[Union['Relationship', 'Unique']]] = {}

        self._col_ref_to_pid_concept: dict['ColumnRef', 'EntityType'] = {}

    def subtype_map(self):
        return self._subtype_map

    def value_type_bindings(self):
        return self._value_type_bindings

    def constructor_bindings(self):
        return self._constructor_bindings

    def referent_constructor_bindings(self):
        return self._referent_constructor_bindings

    def subtype_bindings(self):
        return self._subtype_bindings

    def subtype_bindings_of(self, subtype: Type):
        assert isinstance(subtype, EntityType)
        return self._subtype_bound_thru[subtype]

    def bound_relationships(self):
        return self._bound_relationships

    def bound_external_ucs(self):
        return self._bound_external_ucs

    def referent_bindings(self):
        return self._referent_bindings

    def role_bindings(self):
        return self._bound_role

    def role_bindings_of(self, role: Role):
        return self._role_bound_thru[role]

    def get_ref_scheme_type(self, entity_type: Type):
        assert isinstance(entity_type, EntityType)
        return self._ref_type_of.get(entity_type)

    def get_ctor_binding(self, ref_binding: Binding):
        return self._ref_binding_to_ctor_binding[ref_binding]

    def referenced_concept(self, col_ref: ColumnRef):
        return self._col_ref_to_pid_concept.get(col_ref)

    def analyze(self):
        self._analyze_constraints()
        self._analyze_subtypes()
        potentially_bound_rels = self._analyze_bindings()
        self._analyze_ctor_roles()
        self._analyze_relationships()
        self._analyze_bound_external_ucs()
        self._analyze_bound_relationships(potentially_bound_rels)

        self._process_bound_relationships()
        self._process_referent_bindings()

    def _analyze_subtypes(self):
        # Initialize subtype and supertype maps
        for subtype_arrow in self._model._subtype_arrows:
            parent = subtype_arrow.end
            child = subtype_arrow.start
            self._subtype_map.setdefault(parent, OrderedSet()).add(child)
            self._supertype_map.setdefault(child, OrderedSet()).add(parent)

        self._analyze_subtype_identifiers()

        # Compute closures for subtypes, supertypes, and types
        self._subtype_closure = {}
        self._supertype_closure = {}
        self._type_closure = {}

        for entity in self._subtype_map.keys() | self._supertype_map.keys():
            self._subtype_closure[entity] = self._compute_closure(entity, self._subtype_map)
            self._supertype_closure[entity] = self._compute_closure(entity, self._supertype_map)
            self._type_closure[entity] = self._subtype_closure[entity] | self._supertype_closure[entity]

    def _analyze_subtype_identifiers(self):
        # Find all leaf subtypes
        leaf_types = OrderedSet()
        non_leaf_types = OrderedSet()
        non_leaf_types.update(self._subtype_map.keys())

        for entity in self._supertype_map:
            if entity in self._subtype_map:
                non_leaf_types.add(entity)
            else:
                leaf_types.add(entity)

        # If leaf has an ID relationship then that's what we use or else look up it bottom up and throw if none found
        memo = set()
        for leaf_type in leaf_types:
            self._find_supertype_identifier_bottom_up(leaf_type, memo, [])

        # For any intermediate or root nodes that didn't get an ID relationship during the bottom-up, try that followed
        # by a top-down lookup (may result in multiple ID relationships from different subtypes)
        for non_leaf_type in non_leaf_types:
            if non_leaf_type in memo:
                continue
            try:
                self._find_supertype_identifier_bottom_up(non_leaf_type, memo, [])
            except KeyError:
                # ignore error, since the top-down approach will throw if none exists either
                self._find_subtype_identifiers_top_down(non_leaf_type, memo, [])

    def _find_supertype_identifier_bottom_up(self, entity_type: EntityType, memo: set[EntityType],
                                             lookup_chain: list[str]):
        """
        Looks up an Identifier relationship or a preferred External UC for an entity type bottom up or throws an error
        otherwise.

        It's designed to be used on leaf types. If used on intermediate nodes if they don't have a preferred ID, any
        exception should be swallowed and top-down used to look up a set of possible ID relationships coming from its
        subtypes.
        """
        if entity_type in memo:
            return self._identifiers_of_subtype.get(entity_type)

        if lookup_chain is None:
            lookup_chain = []
        lookup_chain.append(entity_type.display())

        identifier = self._model.identifier_of(entity_type)
        if identifier is not None:
            ids = OrderedSet()
            ids.add(identifier)
            self._identifiers_of_subtype[entity_type] = ids
            memo.add(entity_type)  # use memoization to reduce lookups
            return ids
        # Check supertype
        supertypes = self._supertype_map.get(entity_type)
        if supertypes is None:
            raise KeyError(f'No reference scheme found for {entity_type}, in chain {lookup_chain}')
        if len(supertypes) > 1:
            raise NotImplementedError(f'Multiple supertypes are not supported yet (in {entity_type})')
        for supertype in supertypes:
            ids = self._find_supertype_identifier_bottom_up(supertype, memo, lookup_chain)
            self._identifiers_of_subtype[entity_type] = ids
            memo.add(entity_type)  # use memoization to reduce lookups
            return ids

    def _find_subtype_identifiers_top_down(self, entity_type: EntityType, memo: set[EntityType],
                                           lookup_chain: list[str]):
        if entity_type in memo:
            return self._identifiers_of_subtype.get(entity_type)

        subtypes = self._subtype_map.get(entity_type)
        if subtypes is None:
            raise KeyError(f'No reference scheme found for {entity_type}, in chain {lookup_chain}')

        ids = OrderedSet()
        for subtype in subtypes:
            identifier = self._model.identifier_of(subtype)
            if identifier is not None:
                ids.add(identifier)
            else:
                subtype_ids = self._find_subtype_identifiers_top_down(subtype, memo, lookup_chain)
                ids.update(subtype_ids)

        self._identifiers_of_subtype[entity_type] = ids
        memo.add(entity_type)  # use memoization to reduce lookups
        return ids

    @staticmethod
    def _compute_closure(entity, relation_map):
        closure = OrderedSet()
        stack = [entity]
        while stack:
            current = stack.pop()
            for related in relation_map.get(current, []):
                if related not in closure:
                    closure.add(related)
                    stack.append(related)
        return closure

    def _analyze_bindings(self):
        potentially_bound_rels = {}
        for binding in self._model._bindings:
            role = self._process_binding(binding)
            # if role is None then such binding doesn't bind to a role
            if role is None:
                continue

            if isinstance(binding, ConceptBinding):
                binding = self._concept_binding_to_actual[binding]

            if role not in self._bound_role:
                self._bound_role[role] = []
            self._bound_role[role].append(binding)
            self._categorize_binding(binding, role)

            # Group bindings by relationship and table
            rel = PotentiallyBoundRelationship(role.part_of, binding.column.table)
            if rel not in potentially_bound_rels:
                potentially_bound_rels[rel] = []
            potentially_bound_rels[rel].append(binding)
        return potentially_bound_rels

    def _process_binding(self, binding: 'Binding'):
        """Dispatch binding to the appropriate type-specific handler."""
        if isinstance(binding, ConceptBinding):
            role = self._handle_concept_binding(binding)
        elif isinstance(binding, FilteringSubtypeBinding):
            return self._handle_filtering_binding(binding)
        elif isinstance(binding, SubtypeBinding):
            role = self._handle_subtype_binding(binding)
        elif isinstance(binding, IdentifierBinding):
            role = self._handle_identifier_binding(binding)
        elif isinstance(binding, RoleBinding):
            role = self._handle_role_binding(binding)
        else:
            raise Exception(f'{binding} is yet not supported')

        if role is None:
            raise Exception(f'Unable to lookup binding role for {binding}')
        return role

    def _handle_concept_binding(self, binding: 'ConceptBinding'):
        entity_type = binding.entity_type
        if entity_type in self._inclusive_entity_types:
            actual_binding = IdentifierBinding(binding.column, entity_type=entity_type)
        else:
            actual_binding = SubtypeBinding(binding.column, sub_type=entity_type)
        # Store the actual binding for later reference
        self._concept_binding_to_actual[binding] = actual_binding
        return self._process_binding(actual_binding)

    def _handle_filtering_binding(self, binding: 'FilteringSubtypeBinding'):
        """Process FilteringSubtypeBinding and update filtering and subtype sets."""
        binding = cast(FilteringSubtypeBinding, binding)
        self._filtering_bindings.add(binding)
        self._subtype_bindings.add(binding)
        sub_type = binding.sub_type
        self._ref_type_of[sub_type] = sub_type
        # Initialize list if subtype is encountered for the first time
        if sub_type not in self._subtype_bound_thru:
            self._subtype_bound_thru[sub_type] = []
        self._subtype_bound_thru[sub_type].append(binding)
        return None  # Explicitly return None for filtering bindings

    def _handle_subtype_binding(self, binding: 'SubtypeBinding'):
        """Process SubtypeBinding by looking up ref type and role."""
        ref_types = self._lookup_ref_types_of(binding.sub_type)
        if len(ref_types) > 1:
            raise Exception(f'Multiple ref types are not yet supported, (type: {binding.sub_type})')
        ref_type = ref_types[0]
        role = self._lookup_ctor_role(ref_type)
        self._subtype_binding_references[binding] = role
        self._subtype_bindings.add(binding)
        self._ref_type_of[binding.sub_type] = ref_type
        return role

    def _handle_identifier_binding(self, binding: 'IdentifierBinding'):
        """Process IdentifierBinding by looking up its constructor role."""
        role = self._lookup_ctor_role(binding.entity_type)
        self._constructor_binding_binds_to[binding] = role
        return role

    @staticmethod
    def _handle_role_binding(binding: 'RoleBinding'):
        """Process RoleBinding by directly using the role attribute."""
        return binding.role

    def _lookup_identifiers_of(self, entity_type: EntityType):
        identifier = self._model.identifier_of(entity_type)
        if identifier is not None:
            ids = OrderedSet()
            ids.add(identifier)
            return ids
        ids = self._identifiers_of_subtype.get(entity_type)
        if ids is not None:
            return ids
        raise Exception(f'Cannot look up identifier relationship for {entity_type.display()}')

    def _lookup_ref_types_of(self, entity_type: EntityType):
        ids = self._lookup_identifiers_of(entity_type)
        ref_types = OrderedSet()
        for identifier in ids:
            for role in self._get_identifier_roles(identifier):
                if role in self._constructor_roles:
                    sibling_role = role.sibling()
                    assert sibling_role is not None
                    ref_types.add(sibling_role.player())
                    break
        return ref_types

    @staticmethod
    def _get_identifier_roles(identifier):
        if isinstance(identifier, Relationship) or isinstance(identifier, Unique):
            return identifier.roles()
        else:
            raise ValueError(f'Identifier type {type(identifier)} is not supported')

    def _lookup_ctor_role(self, entity_type: 'EntityType'):
        ids = self._lookup_identifiers_of(entity_type)
        if len(ids) > 1:
            raise NotImplementedError(f'Multiple identifier lookups are not yet supported (type: {entity_type.display()})')

        identifier = ids[0]
        for role in self._get_identifier_roles(identifier):
            player = role.player()
            if player == entity_type or self._is_part_of_type_closure(entity_type, player):
                # ctor role is the only sibling of the constructed entity type role
                return role.sibling()
        raise ValueError(f'Cannot find ctor role for {entity_type}')

    def _is_part_of_type_closure(self, entity_type: EntityType, other: Type):
        closure = self._type_closure.get(entity_type)
        return closure is not None and other in closure

    def _categorize_binding(self, binding: 'Binding', role: 'Role'):
        player = role.player()
        if isinstance(player, AbstractValueType):
            self._value_type_bindings.add(binding)
        elif isinstance(player, EntityType):
            self._entity_type_bindings.add(binding)
        else:
            raise Exception(f'Binding {binding} is not supported')

    def _analyze_constraints(self):
        for constraint in self._model.constraints():
            if isinstance(constraint, Unique):
                self._process_uc(constraint)
            elif isinstance(constraint, Mandatory):
                self._mandatory_roles.update(constraint.roles())

    def _process_uc(self, constraint: 'Unique'):
        if not constraint.is_preferred_identifier:
            return

        roles = constraint.roles()
        self._constructor_roles.update(roles)  # Mark as constructor roles

        num_roles = len(roles)
        if num_roles == 1:
            self._process_internal_uc(constraint)
        elif num_roles > 1:
            self._external_preferred_uc.add(constraint)

    def _process_internal_uc(self, constraint: Unique):
        self._internal_preferred_uc.add(constraint)

        roles = constraint.roles()
        constructor_role = roles[0]
        rel = constructor_role.part_of

        if rel.arity() != 2:
            raise Exception('Invalid Identifier relationship configuration')

        player = constructor_role.player()
        if isinstance(player, AbstractValueType):
            self._identifier_relationships.add(rel)
        elif isinstance(player, EntityType):
            self._entity_identifier_relationships.add(rel)
        else:
            raise Exception(f'Identifier relationship {rel.pprint()} has unsupported player type {player.name()}')

        concept = constructor_role.sibling().player()
        self._entity_type_identifier[concept] = rel
        self._inclusive_entity_types.add(concept)

    def _analyze_ctor_roles(self):
        for constructor_role in self._constructor_roles:
            self._classify_constructor_bindings(constructor_role)

    def _classify_constructor_bindings(self, constructor_role: 'Role'):
        if constructor_role not in self._bound_role:
            return

        for binding in self._bound_role[constructor_role]:
            if isinstance(binding, IdentifierBinding):
                if not binding.column.references:
                    role = self.lookup_binding_role(binding)
                    self._col_ref_to_pid_concept[binding.column.ref()] = role.player()
                    self._constructor_bindings.add(binding)
                else:
                    self._referent_constructor_bindings.add(binding)
            elif isinstance(binding, SubtypeBinding):
                self._referent_constructor_bindings.add(binding)

    def _analyze_relationships(self):
        for relationship in self._model.relationships():
            arity = relationship.arity()
            if arity == 1:
                self._unary_relationships.add(relationship)
            elif arity == 2:
                self._binary_relationships.add(relationship)

    def _analyze_bound_relationships(self, potentially_bound_rels):
        for rel, bindings in potentially_bound_rels.items():
            rel.bindings.extend(bindings)
            self._analyze_bound_relationship(rel)

    def _analyze_bound_relationship(self, meta: 'PotentiallyBoundRelationship'):
        arity = meta.relationship.arity()
        if arity == 1:
            # Handle unary relationships
            if meta.bindings:
                self._bound_relationships.add(meta)
            return
        elif arity == 2:
            self._process_binary_bound_relationship(meta)
        else:
            self._process_nary_bound_relationship(meta)

    def _process_binary_bound_relationship(self, meta: PotentiallyBoundRelationship):
        relationship = meta.relationship
        key_role, value_role = self._identify_roles(relationship)
        key_type, value_type = key_role.player(), value_role.player()
        if not isinstance(key_type, EntityType):
            raise Exception(f'Key role played by {key_type.display()} must be an entity type')
        # TODO : this below should really check if they bound for the same table
        key_role_is_bound = key_role in self._bound_role
        value_role_is_bound = value_role in self._bound_role
        value_role_is_ctor = value_role in self._constructor_roles
        key_role_infers_emap = self._infers_key_role_emap(key_type, meta.table, key_role_is_bound, value_role_is_ctor)
        fully_bound = value_role_is_bound and (key_role_is_bound or key_role_infers_emap)
        is_bound_value_type_relationship = isinstance(value_type, AbstractValueType) and fully_bound
        is_bound = self._is_bound_identifier_relationship(value_role_is_ctor, value_role_is_bound, key_role_is_bound) or \
                   is_bound_value_type_relationship or \
                   self._is_bound_entity_type_relationship(key_type, value_type, key_role, value_role, meta.table)
        if is_bound:
            self._bound_relationships.add(meta)
        else:
            raise Exception(
                f'Bound relationship `{relationship.pprint()}` must have at least one bound role and '
                f'one inferred entity map'
            )

    def _process_nary_bound_relationship(self, meta: PotentiallyBoundRelationship):
        relationship = meta.relationship
        roles = relationship.roles()
        has_entity_type_player = False
        bound_roles = OrderedSet()
        for binding in meta.bindings:
            if not isinstance(binding, RoleBinding):
                raise Exception(f'Every binding in an N-ary relationship must be a RoleBinding, {relationship.pprint()}')
            bound_roles.add(binding.role)
        is_bound = True
        for role in roles:
            if isinstance(role.player(), EntityType):
                has_entity_type_player = True
            if role not in bound_roles:
                is_bound = False
                break
        if not has_entity_type_player:
            raise ValueError(
                f'Bound relationship `{relationship.pprint()}` must have at least one Role played by an entity type'
            )
        if is_bound:
            self._bound_relationships.add(meta)
        else:
            raise Exception(f'Bound relationship `{relationship.pprint()}` must have all roles bound')

    def _identify_roles(self, rel: 'Relationship') -> tuple['Role', 'Role']:
        key_role, value_role = None, None
        for role in rel.roles():
            if isinstance(role.player(), AbstractValueType):
                value_role = role
            # if both roles are entity types, arbitrarily pick one as the key role
            elif key_role is None and role not in self._constructor_roles:
                key_role = role
            else:
                value_role = role
        assert key_role is not None
        assert value_role is not None
        return key_role, value_role

    def _infers_key_role_emap(self, key_type: EntityType, table, key_role_is_bound, value_role_is_ctor):
        if key_role_is_bound or value_role_is_ctor:
            return False
        if key_type.is_composite():
            return self._exists_bound_external_uc(key_type, table)
        return self._exists_ctor_binding(key_type, table) or self._exists_ref_binding(key_type, table) or\
                self._exists_subtype_binding(key_type, table)

    def _exists_bound_external_uc(self, key_type, table):
        for bound_constraint in self._bound_external_ucs:
            if bound_constraint.concept == key_type and bound_constraint.table == table:
                return True
        return False

    def _exists_subtype_binding(self, entity_type, table):
        if entity_type not in self._supertype_map:
            return False

        for binding in self._subtype_bindings:
            if binding.column.table == table and binding.sub_type == entity_type:
                ref_types = self._lookup_ref_types_of(entity_type)
                # TODO: check how this works with multi-table inheritance
                for ref_type in ref_types:
                    if self._exists_ctor_binding(ref_type, table):
                        return True
        return False

    @staticmethod
    def _is_bound_identifier_relationship(value_role_is_ctor, value_role_is_bound, key_role_is_bound):
        if value_role_is_ctor and value_role_is_bound:
            if key_role_is_bound:
                raise Exception('Identifier relationship must not have the key role bound')
            return True
        return False

    def _is_bound_entity_type_relationship(self, key_type, value_type, key_role, value_role, table):
        if isinstance(value_type, EntityType) and isinstance(key_type, EntityType):
            value_role_infers_emap = self._exists_ctor_binding(value_type, table) or self._exists_ref_binding(value_type, table)
            return (
                (key_role in self._bound_role and value_role in self._bound_role) or
                (value_role in self._bound_role and value_role_infers_emap) or
                (key_role in self._bound_role and value_role_infers_emap)
            )
        if isinstance(value_type, AbstractValueType) and isinstance(key_type, AbstractValueType):
            raise Exception('Binary Relationship cannot have more than one ValueType role')
        return False

    def _exists_ctor_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        return self.lookup_ctor_binding(entity_type, table) is not None

    def lookup_ctor_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        # TODO : implement more efficient lookup
        for binding, role in self._constructor_binding_binds_to.items():
            sibling_role = role.sibling()
            assert sibling_role is not None
            if sibling_role.player() == entity_type and binding.column.table == table:
                return binding
        if entity_type in self._subtype_closure:
            for subtype in self._subtype_closure[entity_type]:
                cand = self.lookup_ctor_binding(subtype, table)
                if cand:
                    return cand
        return None

    def _exists_ref_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        return self.lookup_ref_binding(entity_type, table) is not None

    def lookup_ref_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        # TODO : implement more efficient lookup
        for binding, role in self._subtype_binding_references.items():
            if isinstance(binding, SubtypeBinding):
                binding = cast(SubtypeBinding, binding)
                sibling_role = role.sibling()
                assert sibling_role is not None
                ref_type = self._ref_type_of[binding.sub_type]
                if sibling_role.player() == ref_type and binding.column.table == table and entity_type == binding.sub_type:
                    return binding
        if entity_type in self._subtype_closure:
            for subtype in self._subtype_closure[entity_type]:
                cand = self.lookup_ref_binding(subtype, table)
                if cand:
                    return cand
        return None

    def lookup_binding_role(self, binding: 'Binding'):
        if isinstance(binding, IdentifierBinding):
            role = self._constructor_binding_binds_to[binding]
        elif isinstance(binding, RoleBinding):
            role = binding.role
        elif isinstance(binding, FilteringSubtypeBinding):
            role = self._lookup_ctor_role(binding.sub_type)
        elif isinstance(binding, SubtypeBinding):
            role = self._subtype_binding_references[binding]
        else:
            raise Exception(f'Binding {binding} is not supported')
        if role is None:
            raise Exception(f'Unable to lookup binding role for {binding}')
        return role

    def _process_bound_relationships(self):
        for rel_meta in self._bound_relationships:
            if self.is_identifier_relationship(rel_meta.relationship):
                self._process_identifier_relationship(rel_meta)
            else:
                self._process_non_identifier_relationship(rel_meta)

    def is_identifier_relationship(self, relationship):
        return relationship in self._identifier_relationships or \
               relationship in self._entity_identifier_relationships

    def _process_identifier_relationship(self, rel_meta):
        roles = rel_meta.relationship.roles()
        ctor_role, concept_role = self._classify_identifier_roles(roles)
        ctor_role_bindings, concept_role_bindings = self._categorize_bindings(rel_meta.bindings)
        self._update_role_bound_thru(ctor_role, ctor_role_bindings)
        self._update_role_bound_thru(concept_role, concept_role_bindings)

    def _classify_identifier_roles(self, roles):
        ctor_role = next((role for role in roles if role in self._constructor_roles), None)
        concept_role = next((role for role in roles if role != ctor_role), None)
        return ctor_role, concept_role

    def _categorize_bindings(self, bindings):
        ctor_role_bindings = []
        concept_role_bindings = []
        for binding in bindings:
            if binding in self._constructor_bindings:
                ctor_role_bindings.append(binding)
                concept_role_bindings.append(binding)
            elif binding in self._referent_constructor_bindings:
                ctor_role_bindings.append(binding)
                concept_role_bindings.append(binding)
            elif binding in self._subtype_bindings:
                concept_role_bindings.append(binding)
        return ctor_role_bindings, concept_role_bindings

    def _update_role_bound_thru(self, role, bindings):
        if role not in self._role_bound_thru:
            self._role_bound_thru[role] = []
        self._role_bound_thru[role].extend(bindings)

    def _process_non_identifier_relationship(self, rel_meta):
        role_bindings = {}
        for binding in rel_meta.bindings:
            role = self.lookup_binding_role(binding)
            role_bindings.setdefault(role, []).append(binding)
            if isinstance(binding, RoleBinding) and isinstance(binding.role.player(), EntityType):
                self._referent_bindings.add(binding)
        self._role_bound_thru.update(role_bindings)

    def _process_referent_bindings(self):
        for binding in self._referent_bindings:
            ctor_binding = self._lookup_binding_reference(binding)
            self._ref_binding_to_ctor_binding[binding] = ctor_binding

    def _lookup_binding_reference(self, binding):
        concept = binding.role.player()
        # case 1: concept is an entity type with a reference scheme - just look up the ID relationship
        if concept in self._inclusive_entity_types:
            return self._lookup_inclusive_type_binding(concept, binding)
        # case 2: concept is an entity type with a composite reference scheme
        elif concept in self._composite_entity_types:
            # TODO implement composite reference scheme
            raise Exception('Composite reference scheme not supported yet')
        # case 3: concept has no reference scheme, e.g. exclusive entity type
        else:
            # In this case we look up an IdentifierBinding that has the same column as the binding
            # and role player as the same entity type played by the binding. It must be unique,
            # otherwise we raise an exception.
            cand_bindings = []
            ref_types = self._lookup_ref_types_of(concept)
            for ref_type in ref_types:
                cand_binding = self._lookup_inclusive_type_binding(ref_type, binding)
                if cand_binding and cand_binding.column == binding.column and concept in self._type_closure[ref_type]:
                    cand_bindings.append(cand_binding)
            if len(cand_bindings) != 1:
                raise Exception(f'Binding {binding} has no unique reference to an IdentifierBinding')
            else:
                return cand_bindings[0]

    def _lookup_inclusive_type_binding(self, ref_concept: 'EntityType', binding: 'Binding'):
        identifier_relationship = self._model.identifier_of(ref_concept)
        # TODO optimize
        ctor_binding = None
        concept_role = None
        for role in identifier_relationship.roles():
            if role in self._constructor_roles:
                concept_role = role.sibling()
            else:
                concept_role = role
            break
        for cand_binding in self._role_bound_thru[concept_role]:
            if isinstance(cand_binding, IdentifierBinding) and binding.column == cand_binding.column:
                ctor_binding = cand_binding
                break
        return ctor_binding

    def _analyze_bound_external_ucs(self):
        for unique_constraint in self._external_preferred_uc:
            roles = unique_constraint.roles()
            role_bindings_by_table = self._collect_table_role_bindings(roles)
            for table, role_bindings in role_bindings_by_table.items():
                if len(role_bindings) < len(roles):
                    raise Exception(
                        f'External preferred identifier {unique_constraint} must have all roles bound'
                    )
                bound_uc = BoundExternalPreferredUC(unique_constraint, table, role_bindings)
                self._bound_external_ucs.add(bound_uc)

    def _collect_table_role_bindings(self, roles):
        role_bindings_by_table = {}
        for role in roles:
            bindings = self._bound_role.get(role)
            if not bindings:
                continue
            for binding in bindings:
                table = binding.column.table
                role_bindings_by_table.setdefault(table, {}).setdefault(role, []).append(binding)
        return role_bindings_by_table



from relationalai.early_access.metamodel import ir as meta
from relationalai.early_access.metamodel import types
from relationalai.early_access.lqp import ir as lqp
import datetime

def meta_type_to_lqp(typ: meta.Type) -> lqp.RelType:
    if isinstance(typ, meta.UnionType):
        # TODO - this is WRONG! we need to fix the typer wrt overloading
        typ = typ.types.some()

    assert isinstance(typ, meta.ScalarType)

    if types.is_builtin(typ):
        assert not types.is_any(typ), f"unexpected type: {typ}"
        # TODO: just ocompare to types.py
        if typ.name == "Int":
            return lqp.PrimitiveType.INT
        elif typ.name == "Float":
            return lqp.PrimitiveType.FLOAT
        elif typ.name == "String":
            return lqp.PrimitiveType.STRING
        elif typ.name == "Number":
            # TODO: fix this, this is wrong
            return lqp.PrimitiveType.INT
        elif typ.name == "Decimal":
            return lqp.ValueType.DECIMAL
        elif typ.name == "Date":
            return lqp.ValueType.DATE
        elif typ.name == "DateTime":
            return lqp.ValueType.DATETIME
        elif typ.name == "RowId":
            return lqp.PrimitiveType.UINT128
        else:
            raise NotImplementedError(f"Unknown builtin type: {typ.name}")

    assert types.is_entity_type(typ), f"unexpected type: {typ}"
    return lqp.PrimitiveType.UINT128


def type_from_constant(arg: lqp.PrimitiveValue) -> lqp.RelType:
    if isinstance(arg, int):
        return lqp.PrimitiveType.INT
    elif isinstance(arg, float):
        return lqp.PrimitiveType.FLOAT
    elif isinstance(arg, str):
        return lqp.PrimitiveType.STRING
    # TODO: Direct use of date/datetime is not supported in the IR, so this should be
    # rewritten with construct_date.
    elif isinstance(arg, datetime.date):
        return lqp.ValueType.DATE
    else:
        raise NotImplementedError(f"Unknown constant type: {type(arg)}")



"""
Core functionality for the graphlib package.
"""

from relationalai.early_access.builder import Model, Relationship
from relationalai.early_access.builder import Integer, Float
from relationalai.early_access.builder import where, define, count

class GraphCore():
    _model: Model
    # NOTE: _validate_provided_model(), called from __init__, asserts that
    #   `edge_arg._model` is a `Model`, but type checkers fail to pick
    #   that information up, necessitating this `_model: Model` and
    #   the `# type: ignore[Model]` on assignment of `_model` in __init__.

    def __init__(self, *,
            directed: bool,
            weighted: bool,
            edge_arg: Relationship,
            node_arg: Relationship,
        ):
        """
        Abstract base class for graphs, containing logic common to both
        DirectedGraphCore and UndirectedGraphCore (which inherit from this class).
        """
        assert isinstance(directed, bool), "The `directed` argument must be a boolean."
        assert isinstance(weighted, bool), "The `weighted` argument must be a boolean."
        self.directed = directed
        self.weighted = weighted

        self._validate_provided_edge_relationship(edge_arg)
        self._validate_provided_node_relationship(node_arg)
        self._validate_provided_model(edge_arg, node_arg)
        self._model = edge_arg._model # type: ignore[Model]
        self._edge_arg = edge_arg
        self._node_arg = node_arg

        self._define_node_relationship()
        self._define_edge_relationships()
 
        self._define_num_nodes_relationship()
        self._define_num_edges_relationship()

        self._define_neighbor_relationships()
        self._define_count_neighbor_relationships()
        self._define_common_neighbor_relationship()
        self._define_count_common_neighbor_relationship()

        self._define_degree_relationships()


    def _validate_provided_edge_relationship(self, edge_arg):
        assert isinstance(edge_arg, Relationship), \
            "The `edge` argument must be a Relationship."

        edge_arg_field_count = len(edge_arg._fields)
        edge_arg_field_types = [field.type_str for field in edge_arg._fields]

        # Assert expected arity of the `edge` relationship.
        if self.weighted:
            assert edge_arg_field_count == 3, \
                f"For a weighted graph, the `edge` relationship must " \
                f"have width three, whereas it has width {edge_arg_field_count}."
        else: # not self.weighted
            assert edge_arg_field_count == 2, \
                f"For an unweighted graph, the `edge` relationship must " \
                f"have width two, whereas it has width {edge_arg_field_count}."

        # Assert expected types of the `edge` relationship.
        assert edge_arg_field_types[0] == "Integer", \
            f"The `edge` relationship's first value must have type Integer, " \
            f"whereas it has type {edge_arg_field_types[0]}."
        assert edge_arg_field_types[1] == "Integer", \
            f"The `edge` relationship's second value must have type Integer, " \
            f"whereas it has type {edge_arg_field_types[1]}."
        if edge_arg_field_count == 3:
            assert edge_arg_field_types[2] == "Float", \
                f"The `edge` relationship's third value must have type Float, " \
                f"whereas it has type {edge_arg_field_types[2]}."

    def _validate_provided_node_relationship(self, node_arg):
        assert isinstance(node_arg, Relationship), \
            "The `node` argument must be a Relationship."

        node_arg_field_count = len(node_arg._fields)
        node_arg_field_types = [field.type_str for field in node_arg._fields]

        assert node_arg_field_count == 1, \
            f"The `node` relationship must have arity 1, " \
            f"whereas it has arity {node_arg_field_count}."
        assert node_arg_field_types[0] == "Integer", \
            f"The `node` relationship's value must have type Integer, " \
            f"whereas it has type {node_arg_field_types[0]}."

    def _validate_provided_model(self, edge_arg, node_arg):
        assert isinstance(edge_arg._model, Model), (
            "The `edge` relationship must be attached to a model.")
        assert node_arg._model == edge_arg._model, (
            "The `node` and `edge` relationships must belong to the same model. "
            "Please ensure both relationships are defined in the same model.")


    def _define_node_relationship(self):
        """
        Define the self.node relationship, consume the provided
        self._node_arg and self._edge_arg relationships.
        """
        self.node = self._model.Relationship("{node:Integer}")

        node_a, node_b = Integer.ref(), Integer.ref()
        # Populate with any explicitly provided nodes.
        where(self._node_arg(node_a)).define(self.node(node_a))
        # Populate with all nodes implied by edges.
        ((where(self._edge_arg(node_a, node_b, Float)) if self.weighted else
          where(self._edge_arg(node_a, node_b)))
          .define(self.node(node_a), self.node(node_b)))
        # NOTE: In the future, the nodes implied by edges may be guaranteed
        #   by construction to subset the nodes explicitly provided
        #   in the node relationship (if any).

    def _define_edge_relationships(self):
        """
        Define the self.edge and self.weight relationships,
        consuming the provided self._edge_arg relationship.
        """
        self.edge = self._model.Relationship("{node_a:Integer} has edge to {node_b:Integer}")
        self.weight = self._model.Relationship("{node_a:Integer} has edge to {node_b:Integer} with weight {weight:Float}")

        node_a, node_b, edge_weight = Integer.ref(), Integer.ref(), Float.ref()

        if self.directed and self.weighted:
            where(self._edge_arg(node_a, node_b, edge_weight)).define(
                self.weight(node_a, node_b, edge_weight),
                self.edge(node_a, node_b)
            )
        elif self.directed and not self.weighted:
            where(self._edge_arg(node_a, node_b)).define(
                self.weight(node_a, node_b, 1.0),
                self.edge(node_a, node_b)
            )
        elif not self.directed and self.weighted:
            where(self._edge_arg(node_a, node_b, edge_weight)).define(
                self.weight(node_a, node_b, edge_weight),
                self.weight(node_b, node_a, edge_weight),
                self.edge(node_a, node_b),
                self.edge(node_b, node_a)
            )
        elif not self.directed and not self.weighted:
            where(self._edge_arg(node_a, node_b)).define(
                self.weight(node_a, node_b, 1.0),
                self.weight(node_b, node_a, 1.0),
                self.edge(node_a, node_b),
                self.edge(node_b, node_a)
            )

    def _define_num_nodes_relationship(self):
        """Define the self.num_nodes relationship."""
        self.num_nodes = self._model.Relationship("{num_nodes:Integer}")
        define(self.num_nodes(count(self.node(Integer)) | 0))

    def _define_num_edges_relationship(self):
        """Define the self.num_edges relationship."""
        self.num_edges = self._model.Relationship("{num_edges:Integer}")

        node_a, node_b = Integer.ref(), Integer.ref()

        if self.directed:
            define(self.num_edges(count(node_a, node_b, self.edge(node_a, node_b)) | 0))
        elif not self.directed:
            define(self.num_edges(count(node_a, node_b, self.edge(node_a, node_b), node_a <= node_b) | 0))


    def _define_neighbor_relationships(self):
        """Define the self.[in,out]neighbor relationships."""
        self.neighbor = self._model.Relationship("{node_a:Integer} has neighbor {node_b:Integer}")
        self.inneighbor = self._model.Relationship("{node_a:Integer} has inneighbor {node_b:Integer}")
        self.outneighbor = self._model.Relationship("{node_a:Integer} has outneighbor {node_b:Integer}")

        node_a, node_b = Integer.ref(), Integer.ref()
        where(self.edge(node_a, node_b)).define(self.neighbor(node_a, node_b), self.neighbor(node_b, node_a))
        where(self.edge(node_b, node_a)).define(self.inneighbor(node_a, node_b))
        where(self.edge(node_a, node_b)).define(self.outneighbor(node_a, node_b))
        # Note that these definitions happen to work for both
        # directed and undirected graphs due to `edge` containing
        # each edge's symmetric partner in the undirected case.

    def _define_count_neighbor_relationships(self):
        """
        Define the self.count_[in,out]neighbor relationships.
        Note that these relationships differ from corresponding
        [in,out]degree relationships in that they yield empty
        rather than zero absent [in,out]neighbors.
        Primarily for internal consumption.
        """
        self.count_neighbor = self._model.Relationship("{node:Integer} has neighbor count {count:Integer}")
        self.count_inneighbor = self._model.Relationship("{node:Integer} has inneighbor count {count:Integer}")
        self.count_outneighbor = self._model.Relationship("{node:Integer} has outneighbor count {count:Integer}")

        node_a, node_b = Integer.ref(), Integer.ref()
        where(self.neighbor(node_a, node_b)).define(self.count_neighbor(node_a, count(node_b).per(node_a)))
        where(self.inneighbor(node_a, node_b)).define(self.count_inneighbor(node_a, count(node_b).per(node_a)))
        where(self.outneighbor(node_a, node_b)).define(self.count_outneighbor(node_a, count(node_b).per(node_a)))

    def _define_common_neighbor_relationship(self):
        """Define the self.common_neighbor relationship."""
        self.common_neighbor = self._model.Relationship("{node_a:Integer} and {node_b:Integer} have common neighbor {node_c:Integer}")

        node_a, node_b, node_c = Integer.ref(), Integer.ref(), Integer.ref()
        where(self.neighbor(node_a, node_c), self.neighbor(node_b, node_c)).define(self.common_neighbor(node_a, node_b, node_c))

    def _define_count_common_neighbor_relationship(self):
        """Define the self.count_common_neighbor relationship."""
        self.count_common_neighbor = self._model.Relationship("{node_a:Integer} and {node_b:Integer} have common neighbor count {count:Integer}")

        node_a, node_b, node_c = Integer.ref(), Integer.ref(), Integer.ref()
        where(self.common_neighbor(node_a, node_b, node_c)).define(self.count_common_neighbor(node_a, node_b, count(node_c).per(node_a, node_b)))


    def _define_degree_relationships(self):
        """Define the self.[in,out]degree relationships."""
        self.degree = self._model.Relationship("{node:Integer} has degree {count:Integer}")
        self.indegree = self._model.Relationship("{node:Integer} has indegree {count:Integer}")
        self.outdegree = self._model.Relationship("{node:Integer} has outdegree {count:Integer}")

        node, incount, outcount = Integer.ref(), Integer.ref(), Integer.ref()

        where(
            self.node(node),
            _indegree := where(self.count_inneighbor(node, incount)).select(incount) | 0,
        ).define(self.indegree(node, _indegree))

        where(
            self.node(node),
            _outdegree := where(self.count_outneighbor(node, outcount)).select(outcount) | 0,
        ).define(self.outdegree(node, _outdegree))

        if self.directed:
            where(
                self.node(node),
                _indegree := where(self.indegree(node, incount)).select(incount) | 0,
                _outdegree := where(self.outdegree(node, outcount)).select(outcount) | 0,
            ).define(self.degree(node, _indegree + _outdegree))
        elif not self.directed:
            neighcount = Integer.ref()
            where(
                self.node(node),
                _degree := where(self.count_neighbor(node, neighcount)).select(neighcount) | 0,
            ).define(self.degree(node, _degree))



class UniqueNames:
    def __init__(self):
        # Track count of seen names
        self.seen = dict[str, int]()
        # Maps id to unique name
        self.id_to_name = dict[int,str]()

    def get_name(self, name: str) -> str:
        # TODO: Do we need to be threadsafe?
        if name in self.seen:
            self.seen[name] += 1
            id = self.seen[name]
            return f"{name}_{id}"
        else:
            self.seen[name] = 1
            return f"{name}"

    # Get a unique name for the given id. If the id is already in the map, return the
    # existing name. Otherwise, generate a new name using the suggested_name and
    # store it in the map.
    def get_name_by_id(self, id: int, suggested_name:str) -> str:
        if id in self.id_to_name:
            return self.id_to_name[id]

        name = self.get_name(suggested_name)
        self.id_to_name[id] = name
        return name



from __future__ import annotations
from typing import Any, Iterable, Sequence as PySequence, cast, Tuple, Union
from dataclasses import dataclass, field

from relationalai.early_access.metamodel import ir, compiler as c, builtins as bt, types, visitor, typer, helpers, factory as f

from relationalai.early_access.metamodel.rewrite import Splinter, RewriteListTypes, GarbageCollectNodes
from relationalai.early_access.metamodel.visitor import ReadWriteVisitor
from relationalai.early_access.metamodel.util import OrderedSet, group_by, NameCache

from relationalai.early_access.rel import rel, rel_utils as u, rewrite


#--------------------------------------------------
# Compiler
#--------------------------------------------------

class Compiler(c.Compiler):
    def __init__(self):
        super().__init__([
            RewriteListTypes(),
            GarbageCollectNodes(),
            rewrite.CDC(),
            typer.Checker(),
            typer.Typer(),
            GarbageCollectNodes(),
            rewrite.ExtractCommon(),
            rewrite.Flatten(),
            rewrite.QuantifyVars(),
            Splinter()
        ])
        self.model_to_rel = ModelToRel()

    def do_compile(self, model: ir.Model, options:dict={}) -> str:
        return str(self.model_to_rel.to_rel(model, options=options))


@dataclass
class ModelToRel:
    """ Generates Rel from an IR Model, assuming the compiler rewrites were done. """

    relation_name_cache: NameCache = field(default_factory=NameCache)
    rule_name_cache: NameCache = field(default_factory=NameCache)

    # Map a rel variable to one with a different name
    var_map: dict[rel.Var, rel.Var] = field(default_factory=dict)

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------

    def to_rel(self, model: ir.Model, options:dict = {}) -> rel.Program:
        self._register_external_relations(model)
        if options.get("no_declares"):
            return rel.Program(tuple(self._generate_rules(model)))
        else:
            return rel.Program(tuple([
                *self._generate_declares(model),
                *self._generate_rules(model),
            ]))

    #--------------------------------------------------
    # Top level handlers
    #--------------------------------------------------

    def _generate_declares(self, m: ir.Model) -> list[rel.Declare]:
        """
        Generate declare statements for relations declared by the model and:
            - not built-ins
            - not used as an annotation
            - not annotated as external
            - do not start with ^ (for hardcoded Rel constructors)
            - and are never the target of an update
        """
        rw = ReadWriteVisitor()
        m.accept(rw)

        root = cast(ir.Logical, m.root)

        annotations = [anno.relation for anno in visitor.collect_by_type(ir.Annotation, m.root)]
        reads = m.relations - rw.writes(root) - bt.builtin_relations - annotations
        reads = list(filter(lambda r: not r.name.startswith('^') and not helpers.is_external(r), reads))

        declares: list[rel.Declare] = []
        for r in reads:
            if r.name in rel.infix or r.name in u.OPERATORS:
                continue
            # In case parameter name starts with ':' use its name instead of type name
            def requires_name(field: ir.Field):
                if isinstance(field.type, ir.ScalarType):
                    t = field.type
                    if t == types.Symbol:
                        return rel.MetaValue(field.name[1:])
                    else:
                        return rel.Var(name=u.sanitize(field.name.lower()), type=u.rel_typename(t))
                else:
                    return rel.Var(u.sanitize(field.name.lower()))
            head = tuple([requires_name(f) for f in r.fields])

            # Example: declare test(:a, _x0 in Int, _x1 in String) requires true
            declares.append(rel.Declare(
                rel.atom(self._relation_name(r), head),
                True  # `requires true` does not generate any constraints, that affects performance on the RAI side
            ))
        return declares

    def _generate_rules(self, m: ir.Model) -> list[Union[rel.Def, rel.RawSource]]:
        """ Generate rules for the root of this model.

        Assumes the model already was processed such that it contains a root Logical with
        children that are also Logical tasks representing the rules to generate.
        """
        rules: list[Union[rel.Def, rel.RawSource]] = []
        root = cast(ir.Logical, m.root)
        for child in root.body:
            rules.extend(self._generate_rule(cast(ir.Logical, child)))
        return rules

    def _generate_rule(self, rule: ir.Logical) -> list[Union[rel.Def, rel.RawSource]]:
        """ Generate rules for a nested Logical in a model.

        This is for a top-level Logical, under the root Logical.
        """
        # reset the name cache for each rule
        self.rule_name_cache = NameCache()
        effects, other, aggregates, ranks = self._split_tasks(rule.body)
        if not effects or (aggregates and ranks):
            # nothing to generate for this Logical
            return []

        elif len(effects) == 1:
            # a single effect with a body becomes a single rule
            effect = effects[0]

            # deal with raw sources
            if isinstance(effect, ir.Update) and effect.relation == bt.raw_source:
                # TODO: remove this once the type checker checks this.
                assert(len(effect.args) == 2 and isinstance(effect.args[0], str) and isinstance(effect.args[1], str))
                if effect.args[0] != "rel":
                    return []
                return [rel.RawSource(cast(str, effect.args[1]))]
            else:
                args, lookups = self._effect_args(effect)
                if lookups:
                    other.extend(lookups)
                annos = effect.annotations if isinstance(effect, ir.Update) else []
                return [rel.Def(
                    self._effect_name(effect),
                    args,
                    self.generate_logical_body(other, aggregates, ranks),
                    self.generate_annotations(annos)
                )]
        else:
            # currently we can only deal with multiple effects if they are all updates with
            # no body, which is the pattern for inserting hardcoded data.
            if other or aggregates:
                raise NotImplementedError("Body in logical task with multiple effects.")
            if any(isinstance(effect, ir.Output) for effect in effects):
                raise NotImplementedError("Output in logical task with multiple effects.")
            sample = cast(ir.Update, effects[0]).effect
            if any(cast(ir.Update, effect).effect != sample for effect in effects):
                raise NotImplementedError("Different types of effects in logical task.")

            # Group updates by relation name
            relation_groups = group_by(cast(list[ir.Update], effects), lambda e: self._relation_name(e.relation))

            # Process each relation group
            defs = []
            for name, updates in relation_groups.items():
                effects_to_union = []
                for update in updates:
                    update_args, lookups = self._effect_args(update)
                    if update_args:
                        defs.append(
                            rel.Def(
                                name,
                                update_args,
                                self.generate_body_expr(lookups),
                                self.generate_annotations(update.annotations)
                            )
                        )
                    else:
                        effects_to_union.append(update)

                if effects_to_union:
                    update = updates.some()
                    args, lookups = self._effect_args(update)
                    bodies = []
                    for update in effects_to_union:
                        bodies.append(self.handle(update))
                    for lookup in lookups:
                        bodies.append(self.handle(lookup))

                    defs.append(
                        rel.Def(
                            name,
                            args,
                            rel.Union(tuple(bodies)),
                            self.generate_annotations(update.annotations)
                        )
                    )
            return defs

    def generate_logical_body(self, other, aggregates, ranks):
        """ Generate the body of a rule for a Logical that contains these aggregates/ranks
        and other tasks (i.e., no effects)."""

        if aggregates:
            # push the body into the aggregates; this assumes a rewrite pass already
            # prepared the body to contain only what's needed by the aggregates
            exprs = []
            for agg in aggregates:
                # The variables declared in the relational abstraction are the agg's "projection" + "over"
                abs_vars = OrderedSet.from_iterable(agg.projection)
                result = []
                for arg in agg.args:
                    if helpers.is_aggregate_input(arg, agg):
                        new_arg = arg if isinstance(arg, ir.Var) else self.handle(arg)
                        abs_vars.add(new_arg)
                    else:
                        result.append(self.handle(arg))

                old_var_map = self.var_map
                self.var_map = {}

                common_vars = OrderedSet.from_iterable(agg.projection) & agg.group
                abs_body_exprs = []
                for v in common_vars:
                    orig_rel_v = self.handle_var(v)
                    inner_rel_v = rel.Var("_t" + orig_rel_v.name)
                    self.var_map[orig_rel_v] = inner_rel_v
                    eq_expr = rel.BinaryExpr(orig_rel_v, "=", inner_rel_v)
                    abs_body_exprs.append(eq_expr)

                abs_head = self.handle_list(tuple(abs_vars))
                abs_body = self.generate_body_expr(other)
                if abs_body_exprs:
                    abs_body = rel.create_and([abs_body, *abs_body_exprs])
                rel_abstraction = rel.RelationalAbstraction(abs_head, abs_body)

                self.var_map = old_var_map

                exprs.append(rel.atom(
                    u.rel_operator(agg.aggregation.name),
                    tuple([ rel_abstraction, *result ])
                ))
            return exprs[0] if len(exprs) == 1 else rel.create_and(exprs)
        elif ranks:
            # push the body into the aggregates; this assumes a rewrite pass already
            # prepared the body to contain only what's needed by the aggregates
            exprs = []
            for rank in ranks:
                if all(o for o in rank.arg_is_ascending):
                    rel_name = "std::common::sort"
                elif all(not o for o in rank.arg_is_ascending):
                    rel_name = "std::common::reverse_sort"
                else:
                    raise Exception("Mixed orderings in rank are not supported yet.")

                # The keys (projection) have to be present to preserve bag semantics, but
                # should not affect the ranking. Thus they have to go at the end of the list,
                # and then be ignored in the result.
                old_var_map = self.var_map
                self.var_map = {}
                abs_vars = []
                # Rename the sorted vars to avoid conflicts with the result vars.
                for ir_v in rank.args:
                    orig_rel_v = self.handle_var(ir_v)
                    inner_rel_v = rel.Var("_t" + orig_rel_v.name)
                    abs_vars.append(inner_rel_v)
                    self.var_map[orig_rel_v] = inner_rel_v
                # No need to rename the projection vars since they are ignored in the result.
                abs_vars.extend(tuple([self.handle_var(v) for v in rank.projection]))

                abs_body = self.generate_body_expr(other)
                abstraction = rel.RelationalAbstraction(tuple(abs_vars), abs_body)
                self.var_map = old_var_map

                out_vars = [self.handle_var(v) for v in rank.args]
                key_vars = [rel.Var("_")] * len(rank.projection)
                params = tuple([abstraction, self.handle_var(rank.result), *out_vars, *key_vars])
                exprs.append(rel.atom(rel_name, params))

            return exprs[0] if len(exprs) == 1 else rel.create_and(exprs)
        else:
            # no aggregates or ranks, just return an expression for the body
            return self.generate_body_expr(other)

    def generate_body_expr(self, tasks: list[ir.Task]):
        """ Helper to generate the an expression from the tasks, wrapping in Ands if necessary. """
        if not tasks:
            return True
        elif len(tasks) == 1:
            return self.handle(tasks[0])
        else:
            return rel.create_and([self.handle(b) for b in tasks])

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    def handle(self, n: ir.Node):
        """ Dispatch to the appropriate ir.Node handler. """
        if isinstance(n, ir.PyValue):
            return self.handle_value(type(n), n)
        handler = getattr(self, f"handle_{n.kind}", None)
        if handler:
            return handler(n)
        else:
            raise Exception(f"Rel Compiler handler for '{n.kind}' node not implemented.")

    def handle_list(self, n: Iterable[ir.Node]):
        """ Dispatch each node to the appropriate ir.Node handler. """
        return tuple([self.handle(x) for x in n])

    def handle_value(self, type: Any, value: Any) -> Union[rel.Primitive, rel.MetaValue, rel.Var]:
        """ Handle the value (Node or Value) and wrap in a Metavalue if the type is Symbol. """
        # only handle if it is a Node (e.g. ir.Var or ir.Literal)
        v = self.handle(value) if isinstance(value, ir.Node) else value
        # only wrap if v is a primitive (i.e. not a metavalue or a var, for example).
        if type == types.Symbol and isinstance(v, (str, int, float, bool)):
            return rel.MetaValue(v)
        else:
            return v

    def handle_value_list(self, types, values) -> Tuple[Union[rel.Primitive, rel.MetaValue, rel.Var], ...]:
        return tuple([self.handle_value(t, v) for t, v in zip(types, values)])

    #
    # DATA MODEL
    #
    def handle_scalartype(self, n: ir.ScalarType):
        return n.name
    # TODO - what to generate for other kinds of types?




    #
    # TASKS
    #
    def handle_logical(self, n: ir.Logical) -> rel.Expr:
        # Generate nested expressions for a nested logical
        effects, other, aggregates, ranks = self._split_tasks(n.body)
        if effects:
            raise Exception("Cannot process nested logical with effects.")
        elif aggregates and ranks:
            raise Exception("Cannot process nested logical with both aggregates and ranks.")
        return self.generate_logical_body(other, aggregates, ranks)

    def handle_union(self, n: ir.Union) -> rel.Expr:
        # Generate nested expressions for a nested logical
        body:list[rel.Expr] = []
        for t in n.tasks:
            body.append(self.handle(t))
        return rel.Or(OrderedSet.from_iterable(body))

    def handle_rank(self, n: ir.Rank) -> rel.Expr:
        return rel.atom("rank", tuple([self.handle_var(n.result)]))

    #
    # LOGICAL QUANTIFIERS
    #
    def handle_not(self, n: ir.Not):
        return rel.Not(self.handle(n.task))

    def handle_exists(self, n: ir.Exists):
        vars = self._remove_wildcards(n.vars)
        if vars:
            return rel.Exists(
                self.handle_vars(vars),
                self.handle(n.task)
            )
        else:
            # all vars are wildcards, no need for exists
            return self.handle(n.task)

    #
    # ITERATION
    #

    #
    # RELATIONAL OPERATIONS
    #

    def handle_vars(self, vars: Tuple[ir.Var, ...]):
        return tuple([self.handle_var(v) for v in vars])

    def handle_var(self, n: ir.Var):
        name = n.name if (n.name == "_" or n.name.startswith(':')) else f"_{self._var_name(n)}"
        v = rel.Var(name)
        return self.var_map.get(v, v)

    def handle_literal(self, n: ir.Literal):
        return self.handle_value(n.type, n.value)

    def handle_data(self, n: ir.Data):
        return rel.Atom(rel.Union(tuple([self.handle_value(None, d) for d in n])), tuple([self.handle(d) for d in n.vars]))

    def generate_annotations(self, annos: Iterable[ir.Annotation]):
        """ Helper to cast the handling of ir.Annotations into a tuple of rel.Annotations. """
        return cast(Tuple[rel.Annotation, ...], self.handle_list(annos))

    def handle_annotation(self, n: ir.Annotation):
        # we know that annotations won't have vars, so we can ignore that type warning
        return rel.Annotation(
            n.relation.name,
            self.handle_value_list(self._relation_types(n.relation), n.args) # type:ignore
        )

    def handle_update(self, n: ir.Update):
        return rel.atom(
            self._relation_name(n.relation),
            self.handle_value_list(self._relation_types(n.relation), n.args)
        )

    def handle_lookup(self, n: ir.Lookup):
        # only translate names to Rel operators if the relation is a built-in
        name = self._relation_name(n.relation)
        if bt.is_builtin(n.relation):
            name = u.rel_operator(name)
        types = self._relation_types(n.relation)
        return rel.atom(name, self.handle_value_list(types, n.args))

    def handle_construct(self, n: ir.Construct):
        args = [self.handle_value(None, v) for v in n.values] + [self.handle(n.id_var)]
        return rel.atom("rel_primitive_hash_tuple_uint128", tuple(args))

    #--------------------------------------------------
    # Helpers
    #--------------------------------------------------

    def _relation_types(self, relation: ir.Relation):
        return [f.type for f in relation.fields]

    def _relation_name(self, relation: ir.Relation):
        if helpers.is_external(relation) or bt.is_builtin(relation):
            return relation.name
        else:
            return self.relation_name_cache.get_name(relation.id, relation.name)

    def _var_name(self, var: ir.Var):
        if var.name == "_":
            return "_"
        return self.rule_name_cache.get_name(var.id, u.sanitize(var.name.lower()))

    def _remove_wildcards(self, vars: tuple[ir.Var, ...]):
        return tuple(filter(lambda v: v.name != "_", vars))

    def _register_external_relations(self, model: ir.Model):
        # force all external relations to get a name in the cache, so that internal relations
        # cannot use those names in _relation_name
        for r in model.relations:
            if helpers.is_external(r):
                self.relation_name_cache.get_name(r.id, r.name)

    def _split_tasks(self, tasks: PySequence[ir.Task]) -> tuple[list[Union[ir.Update, ir.Output]], list[ir.Task], list[ir.Aggregate], list[ir.Rank]]:
        effects = []
        aggregates = []
        other_body = []
        ranks = []
        for task in tasks:
            if isinstance(task, (ir.Update, ir.Output)):
                effects.append(task)
            elif isinstance(task, ir.Aggregate):
                aggregates.append(task)
            elif isinstance(task, ir.Rank):
                ranks.append(task)
            else:
                other_body.append(task)
        return effects, other_body, aggregates, ranks


    def _effect_name(self, n: ir.Task):
        """ Return the name to be used for the effect (e.g. the relation name, output, etc). """
        if isinstance(n, ir.Output) and bt.export_annotation in n.annotations:
            return "Export_Relation"
        elif isinstance(n, ir.Output):
            return "output"
        elif isinstance(n, ir.Update):
            return self._relation_name(n.relation)
        else:
            raise Exception(f"Cannot retrieve effect name from node {type(n)}")

    def _effect_args(self, n: ir.Task) -> Tuple[Tuple[Any], list[ir.Task]]:
        """
            Return the arguments for the head of an effect rule and a list of lookups to add
            to the body of the rule.

            The lookups may be necessary because Rel does not allow "missing" in the head,
            so we create a new variable, set the variable to missing in the body (the
            lookup) and use the variable in the head.

            E.g. output(None) becomes output(x): { x = missing }
        """
        orig_args = []
        handled_args = []
        if isinstance(n, ir.Output):
            # TODO - deal with output aliases
            for _, v in n.aliases:
                orig_args.append(v)
                handled_args.append(self.handle_value(None, v))
        elif isinstance(n, ir.Update):
            orig_args.extend(n.args)
            handled_args.extend(self.handle_value_list(self._relation_types(n.relation), n.args))
        else:
            raise Exception(f"Cannot retrieve effect params from node {type(n)}")

        args, lookups = [], []
        for idx, handled in enumerate(handled_args):
            if handled is None:
                var = ir.Var(types.Any, "head")
                args.append(self.handle(var))
                lookups.append(f.lookup(bt.eq, [var, handled]))
            elif not isinstance(handled, rel.Var):
                args.append(handled)
            else:
                # Count how many times this argument has been seen before
                cnt = handled_args[:idx].count(handled)
                if cnt == 0:
                    args.append(handled)
                    continue
                # Deduplicate variable
                new_var = ir.Var(types.Any, handled.name + "_dup" + str(cnt))
                args.append(self.handle(new_var))
                lookups.append(f.lookup(bt.eq, [new_var, orig_args[idx]]))
        return tuple(args), lookups



from abc import abstractmethod
from typing import Optional

from relationalai.early_access.dsl.core.types.unconstrained import UnconstrainedValueType


class BindableTable:
    """
    A class representing a bindable table.
    """

    _table_name: str

    def __init__(self, name: str):
        self._table_name = name

    @property
    def table_name(self):
        return self._table_name

    @abstractmethod
    def key_type(self) -> UnconstrainedValueType:
        pass

    @abstractmethod
    def physical_name(self) -> str:
        pass


class BindableAttribute:

    @property
    @abstractmethod
    def table(self) -> BindableTable:
        pass

    @abstractmethod
    def physical_name(self) -> str:
        pass

    @abstractmethod
    def type(self) -> UnconstrainedValueType:
        pass

    @abstractmethod
    def decimal_scale(self) -> Optional[int]:
        pass

    @abstractmethod
    def decimal_size(self) -> Optional[int]:
        pass



import typing

from relationalai.early_access.dsl.orm.constraints import Unique, Mandatory, RoleValueConstraint, \
    InclusiveSubtypeConstraint, ExclusiveSubtypeConstraint, Range
from relationalai.early_access.dsl.orm.models import Model
from relationalai.early_access.dsl.orm.relationships import Role
from relationalai.early_access.dsl.orm.types import ValueType, EntityType, Concept
from relationalai.early_access.metamodel.util import Printer as BasePrinter, OrderedSet
import relationalai.early_access.builder as qb

class Printer(BasePrinter):
    _unique_constraints: OrderedSet[Unique]
    _mandatory_constraints: OrderedSet[Mandatory]
    _role_value_constraints: OrderedSet[RoleValueConstraint]
    _inclusive_subtype_constraints: OrderedSet[InclusiveSubtypeConstraint]
    _exclusive_subtype_constraints: OrderedSet[ExclusiveSubtypeConstraint]

    def __init__(self, io: typing.Optional[typing.IO[str]] = None):
        # Set the base class field (frozen)
        object.__setattr__(self, 'io', io)

        # Set all mutable fields using object.__setattr__
        object.__setattr__(self, '_unique_constraints', OrderedSet())
        object.__setattr__(self, '_mandatory_constraints', OrderedSet())
        object.__setattr__(self, '_role_value_constraints', OrderedSet())
        object.__setattr__(self, '_inclusive_subtype_constraints', OrderedSet())
        object.__setattr__(self, '_exclusive_subtype_constraints', OrderedSet())


    def to_string(self, model: Model) -> None:
        self._process_constraints(model)

        self._print_nl("import relationalai.early_access.builder as qb")
        self._print_nl("from relationalai.early_access.dsl.orm.models import Model")
        self._print_nl("from relationalai.early_access.dsl.orm.constraints import Range")
        self._nl()

        self._print_nl(f"model = Model(name='{model.name}', is_primary={model.is_primary}"
                       f"{self._print_if_true('dry_run', model.qb_model()._dry_run)})")
        self._print_nl("Concept = model.Concept")
        self._print_nl("ValueType = model.ValueType")
        self._print_nl("EntityType = model.EntityType")
        self._print_nl("Relationship = model.Relationship")
        self._print_nl("Unique = model.unique")
        self._print_nl("Mandatory = model.mandatory")
        self._print_nl("RoleValueConstraint = model.role_value_constraint")
        self._print_nl("InclusiveSubtypeConstraint = model.inclusive_subtype_constraint")
        self._print_nl("ExclusiveSubtypeConstraint = model.exclusive_subtype_constraint")

        self._nl()
        self._handle_value_types(model)

        self._nl()
        self._handle_entity_types(model)

        self._nl()
        self._handle_relationships(model)

        self._nl()
        self._handle_ref_schemes(model)

        self._nl()
        self._handle_uniqueness_constraints()

        self._nl()
        self._handle_mandatory_constraints()

        self._nl()
        self._handle_role_value_constraints()

        self._nl()
        self._handle_subtype_constraints()

    def _handle_value_types(self, model: Model) -> None:
        for name, vt in model.value_types_map().items():
            extends = vt._extends
            base_class = "ValueType" if isinstance(vt, ValueType) else "Concept"

            extends_str = f", extends=[{self._get_type(extends[0])}]" if extends else ""
            self._print_nl(f"{name} = {base_class}('{name}'{extends_str})")

    def _handle_entity_types(self, model: Model) -> None:
        for name, et in model.entity_types_map().items():
            extends = et._extends

            extends_elements = [f"{self._get_type(ext)}" for ext in extends]
            extends_str = f", extends=[{', '.join(extends_elements)}]" if extends else ""
            if isinstance(et, EntityType):
                self._print_nl(f"{name} = EntityType('{name}'{extends_str}"
                               f"{self._print_if_not_empty('refmode', et._ref_schema_name())})")
            else:
                self._print_nl(f"{name} = Concept('{name}'{extends_str})")

    def _handle_relationships(self, model: Model) -> None:
        for rel in model.relationships():
            for r in rel._relations:
                self._print_nl(f"{self._get_relationship_name(r)} = "
                               f"Relationship('{r._madlib}'{self._print_if_not_empty('relation_name', r._passed_short_name)})")

    def _handle_ref_schemes(self, model: Model) -> None:
        for concept_name, concept in model.entity_types_map().items():
            if not isinstance(concept, Concept) or not concept._reference_schemes:
                continue

            for ref_scheme in concept._reference_schemes:
                rel_names = [self._get_relationship_name(rel) for rel in ref_scheme]
                self._print_nl(f"{concept_name}.identify_by({', '.join(rel_names)})")

    def _handle_uniqueness_constraints(self) -> None:
        for constraint in self._unique_constraints:
            if not constraint.is_preferred_identifier:
                elements = []
                for role in constraint.roles():
                    elements.append(self._get_role_name(role))
                self._print_nl(f"Unique([{', '.join(elements)}])")

    def _handle_mandatory_constraints(self):
        for constraint in self._mandatory_constraints:
            role = constraint.roles()[0]
            role_name = self._get_role_name(role)
            self._print_nl(f"Mandatory({role_name})")

    def _handle_role_value_constraints(self) -> None:
        for constraint in self._role_value_constraints:
            role = constraint.roles()[0]
            elements = []

            for value in constraint.values():
                if isinstance(value, Range):
                    if value._start is not None and value._end is not None:
                        elements.append(f"Range.between({value._start}, {value._end})")
                    elif value._end is not None:
                        elements.append(f"Range.to_value({value._end})")
                    else:
                        elements.append(f"Range.from_value({value._start})")
                elif isinstance(value, str):
                    elements.append(repr(value))
                else:
                    elements.append(str(value))

            role_name = self._get_role_name(role)
            self._print_nl(f"RoleValueConstraint({role_name}, [{', '.join(elements)}])")

    def _handle_subtype_constraints(self):
        self._emit_subtype_constraints(self._inclusive_subtype_constraints, "InclusiveSubtypeConstraint")
        self._nl()
        self._emit_subtype_constraints(self._exclusive_subtype_constraints, "ExclusiveSubtypeConstraint")

    def _emit_subtype_constraints(self, constraints, constraint_type: str):
        for constraint in constraints:
            concepts = [self._get_type(c) for c in constraint.concepts()]
            self._print_nl(f"{constraint_type}({', '.join(concepts)})")

    def _process_constraints(self, model: Model) -> None:
        for c in model.constraints():
            if isinstance(c, Unique):
                self._unique_constraints.add(c)
            elif isinstance(c, Mandatory):
                self._mandatory_constraints.add(c)
            elif isinstance(c, RoleValueConstraint):
                self._role_value_constraints.add(c)
            elif isinstance(c, InclusiveSubtypeConstraint):
                self._inclusive_subtype_constraints.add(c)
            elif isinstance(c, ExclusiveSubtypeConstraint):
                self._exclusive_subtype_constraints.add(c)

    def _get_role_name(self, role: Role) -> str:
        return f"{self._get_relationship_name(role._relationship)}[{role._concept._name}]"

    @staticmethod
    def _get_relationship_name(rel: qb.Relationship) -> str:
        if rel._parent and rel._short_name:
            parent_name = rel._parent._name
            prefix = "qb." if parent_name in Concept.builtins else ""
            return f"{prefix}{parent_name}.{rel._short_name}"
        return rel._name

    @staticmethod
    def _get_type(t: Concept) -> str:
        name = 'Integer' if t._name == 'Int' else t._name
        return f"qb.{name}" if name in Concept.builtins else name

    @staticmethod
    def _print_if_not_empty(label: str, value: typing.Optional[str]) -> str:
        return f", {label}='{value}'" if value else ""

    @staticmethod
    def _print_if_true(label: str, value: bool) -> str:
        return f", {label}=True" if value else ""


#--------------------------------------------------
# Compiler
#--------------------------------------------------

from relationalai import debugging
from relationalai.early_access.lqp import ir as lqp, llqp
from relationalai.early_access.lqp.passes import lqp_passes
from relationalai.early_access.lqp.model2lqp import to_lqp as ir_to_lqp
from relationalai.early_access.metamodel import ir, compiler as c

class Compiler(c.Compiler):
    def __init__(self):
        super().__init__(lqp_passes())

    def do_compile(self, model: ir.Model, options:dict={}) -> str:
        lqp_ir = ir_to_lqp(model)
        self._lqp_result = lqp_ir
        llqp_config = options.get("llqp_config", {})
        lqp_str = llqp.program_to_llqp(lqp_ir, llqp_config)

        with debugging.span("passes") as span:
            with debugging.span("LQP Emitter") as span:
                span["metamodel"] = llqp.program_to_llqp(lqp_ir, llqp.ugly_config)

        return lqp_str

    def get_lqp_ir(self) -> lqp.LqpProgram:
        """ Returns the LQP IR generated by the compiler. """
        assert self._lqp_result is not None, "LQP IR not generated yet"
        return self._lqp_result



from typing import Any, Optional

from relationalai.early_access.builder import Concept as QBConcept
from relationalai.early_access.builder.builder import python_types_to_concepts, Relationship


class Concept(QBConcept):

    def __init__(self, model, name: str, extends: list[Any] = [], identify_by:dict[str, Any]={}):
        self._dsl_model = model
        # create an orm Relationship for each Concept to be able to refer to them in DSL model
        identify_args = {}
        if identify_by:
            for k, v in identify_by.items():
                if python_types_to_concepts.get(v):
                    v = python_types_to_concepts[v]
                if isinstance(v, Concept):
                    identify_args[k] = self._dsl_model.Relationship(f"{{{name}}} has {{{k}:{v._name}}}", relation_name=k)
                elif isinstance(v, Relationship):
                    identify_args[k] = v
                else:
                    raise ValueError(f"identify_by must be either a Concept or Relationship: {k}={v}")
            # add constraints required for reference schema
            self._dsl_model._ref_scheme_constraints(*identify_args.values())
        super().__init__(name, extends, model.qb_model(), identify_args)

    def identify_by(self, *relations:Relationship):
        super().identify_by(*relations)
        self._dsl_model._ref_scheme_constraints(*relations)

    def _is_value_type(self) -> bool:
        if len(self._extends) == 1:
            ext_concept = self._extends[0]
            if ext_concept._is_primitive():
                return True
            else:
                return ext_concept._is_value_type()
        return False

    def __eq__(self, other):
        return isinstance(other, Concept) and self._name == other._name

    def __hash__(self):
        return hash(self._name)


class EntityType(Concept):

    def __init__(self, model, nm, extends: list[Concept] = [], ref_schema_name: Optional[str] = None):
        self._domain = extends
        super().__init__(model, nm, extends)
        self.__ref_schema_name = ref_schema_name

    def _qualified_name(self):
        return self._name

    def _is_composite(self):
        return len(self._domain) > 1

    def _ref_schema_name(self):
        return self.__ref_schema_name

    def _is_value_type(self) -> bool:
        return False


class ValueType(Concept):

    def __init__(self, model, nm, extends: Optional[Concept] = None):
        super().__init__(model, nm, [extends] if extends is not None else [])

    def _is_value_type(self) -> bool:
        return True


# pyright: reportArgumentType=false
from itertools import product
from types import MappingProxyType
from typing import cast, Union, Optional

from relationalai.early_access.dsl.bindings.tables import (Binding, IdentifierBinding, RoleBinding,
                                                           FilteringSubtypeBinding, SubtypeBinding, BindableColumn,
                                                           DEFAULT_DECIMAL_SCALE, DEFAULT_DECIMAL_SIZE)
from relationalai.early_access.dsl.codegen.common import PotentiallyBoundRelationship, BoundExternalPreferredUC
from relationalai.early_access.dsl.codegen.reasoner import Reasoner
from relationalai.early_access.dsl.codegen.relations import EntityMap, ValueMap, EntitySubtypeMap, RoleMap, \
    CompositeEntityMap
from relationalai.early_access.dsl.core import std
from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.core.relations import rule, AbstractRelation, Relation, EntityInstanceRelation
from relationalai.early_access.dsl.core.rules import Annotation, Vars
from relationalai.early_access.dsl.core.types import AbstractValueType
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.core.types.constrained.nominal import ValueType
from relationalai.early_access.dsl.core.types.standard import RowId, Decimal, Integer, Symbol, Float
from relationalai.early_access.dsl.core.utils import camel_to_snake
from relationalai.early_access.dsl.ontologies.relationships import Reading
from relationalai.early_access.dsl.ontologies.roles import Role
from relationalai.early_access.dsl.types.entities import EntityType
from relationalai.early_access.metamodel.util import OrderedSet


DEFAULT_WEAVER_CONFIG = MappingProxyType({
    'decimal_size': DEFAULT_DECIMAL_SIZE,
    'decimal_scale': DEFAULT_DECIMAL_SCALE,
})


class Weaver:

    def __init__(self, model, config=None):
        self._model = model
        self._config = config or DEFAULT_WEAVER_CONFIG
        self._reasoner = Reasoner(model)

        # Physical Relations & Classification
        self._relations: dict[str, 'AbstractRelation'] = {}
        self._preferred_id_relations: OrderedSet['AbstractRelation'] = OrderedSet()
        self._identifier_relations: OrderedSet['Relation'] = OrderedSet()
        self._entity_map_relations: OrderedSet['Relation'] = OrderedSet()
        self._subtype_entity_map_relations: OrderedSet['AbstractRelation'] = OrderedSet()
        self._value_map_relations: OrderedSet['Relation'] = OrderedSet()
        self._entity_population_relations: OrderedSet['Relation'] = OrderedSet()

        self._binding_to_value_map: dict['Binding', 'ValueMap'] = {}
        self._constructor_binding_to_value_map: dict['Binding', 'EntityMap'] = {}
        self._entity_to_id_relations: dict['EntityType', 'AbstractRelation'] = {}
        self._concept_to_identifies_relations: dict['Type', 'AbstractRelation'] = {}
        self._subtype_to_entity_map: dict['EntityType', list['EntitySubtypeMap']] = {}
        self._composite_entity_to_entity_maps: dict['EntityType', list['CompositeEntityMap']] = {}

    def generate(self):
        self._reasoner.analyze()
        self._generate_value_maps()
        self._generate_entity_maps()
        self._generate_composite_entity_maps()
        self._generate_subtype_entity_maps()
        self._generate_entity_populations()
        self._generate_non_identifier_relationships()

    def _value_converter(self, col: BindableColumn) -> Optional[AbstractRelation]:
        if col.type() != Decimal:
            return None

        size = col.decimal_size()
        scale = col.decimal_scale()
        if size is None or scale is None:
            raise ValueError(f'Decimal column {col.physical_name()} must have size and scale defined')

        if size == self._cfg_decimal_size() and scale == self._cfg_decimal_scale():
            return None

        return self._ref_decimal_convert(size, scale)

    def _ref_decimal_convert(self, size: int, scale: int):
        fqn = self._fqn_decimal_convert(size, scale)
        rel = self._relations.get(fqn)
        if rel is None:
            rel = self._model.external_relation(fqn, Decimal, Decimal)
            unpack_decimal = std.unpack(Decimal, Symbol, Symbol, Symbol, Integer)
            with rel:
                @rule(Annotation.INLINE)
                def decimal_convert(raw, rez):
                    name, t_bits, t_decimals, int_val, bits, decimals, div, float_rez = Vars(
                        Symbol, Symbol, Symbol, Integer, Integer, Integer, Integer, Float
                    )
                    unpack_decimal(raw, name, t_bits, t_decimals, int_val)
                    std.mirror_lower(t_bits, bits)
                    std.mirror_lower(t_decimals, decimals)
                    std.power(10, decimals, div)
                    _= float_rez == int_val / div
                    std.new_decimal(self._cfg_decimal_size(), self._cfg_decimal_scale(), float_rez, rez)

            self._relations[fqn] = rel
        return rel

    def _fqn_decimal_convert(self, size: int, scale: int):
        return f'decimal_{size}_{scale}_to_{self._cfg_decimal_size()}_{self._cfg_decimal_scale()}_convert'

    def _generate_value_maps(self):
        for binding in self._reasoner.value_type_bindings():
            if not isinstance(binding, FilteringSubtypeBinding):
                self._gen_value_map(binding)

    def _generate_entity_maps(self):
        for binding in self._reasoner.constructor_bindings():
            self._gen_entity_map(binding)
            self._gen_preferred_id(binding)
        for binding in self._reasoner.referent_constructor_bindings():
            if not isinstance(binding, FilteringSubtypeBinding):
                self._gen_entity_map(binding)

    def _generate_composite_entity_maps(self):
        for bound_constraint in self._reasoner.bound_external_ucs():
            entity_map_combinations = self._get_entity_map_combinations_for_composite(bound_constraint)

            composite_emaps = []
            fqn = self._fqn_composite_entity_map(bound_constraint)
            concept = bound_constraint.concept
            for idx, entity_map_combination in enumerate(entity_map_combinations):
                rel_nm = f'{fqn}{idx}' if idx > 0 else fqn
                composite_emap = CompositeEntityMap(Namespace.top, rel_nm, RowId, concept, *entity_map_combination)
                composite_emaps.append(composite_emap)
                self._register_relation(fqn, composite_emap, self._entity_map_relations)
            self._composite_entity_to_entity_maps[concept] = composite_emaps

    def _get_entity_map_combinations_for_composite(self, bound_constraint: BoundExternalPreferredUC) -> list[list['EntityMap']]:
        # Start with the role bindings to build the combinations
        role_bindings = bound_constraint.role_bindings
        # Build a list of binding lists for each role in the right order
        role_binding_lists = [role_bindings[role] for role in bound_constraint.constraint.roles()]
        # Generate all possible combinations of bindings and convert each combination to a list
        all_combinations = [list(combination) for combination in product(*role_binding_lists)]  # pyright: ignore[reportCallIssue]
        # Construct entity map combinations using the _ref_entity_map method
        entity_map_combinations = [
            [self._ref_entity_map(binding, binding.role) for binding in combination]
            for combination in all_combinations
        ]
        return entity_map_combinations

    @staticmethod
    def _fqn_composite_entity_map(bound_constraint: BoundExternalPreferredUC):
        concept = bound_constraint.concept
        assert isinstance(concept, EntityType)
        concept_nm = camel_to_snake(concept.display())
        source_nm = bound_constraint.table.physical_name()
        return f'{concept_nm}__impl__{source_nm}_row_to_{concept_nm}'

    def _generate_subtype_entity_maps(self):
        for binding in self._reasoner.subtype_bindings():
            self._gen_subtype_entity_map(binding)

    def _generate_entity_populations(self):
        for entity_type, relation in self._entity_to_id_relations.items():
            self._gen_entity_population(entity_type, relation)
        for entity_type, relations in self._composite_entity_to_entity_maps.items():
            for relation in relations:
                self._gen_entity_population(entity_type, relation)
        for subtype, relations in self._subtype_to_entity_map.items():
            for relation in relations:
                self._gen_entity_population(subtype, relation)
        # add union population relations for those missing one
        for parent_type, subtypes in self._reasoner.subtype_map().items():
            self._gen_union_entity_population(parent_type, subtypes)

    def _gen_union_entity_population(self, parent_type: EntityType, subtypes: OrderedSet[EntityType]):
        has_no_identifier = self._model.identifier_of(parent_type) is None
        has_no_entity_map = parent_type not in self._subtype_to_entity_map
        if has_no_identifier and has_no_entity_map:
            for subtype in subtypes:
                with parent_type:
                    @rule()
                    def add_subtype_population(val):
                        subtype(val)

    def _generate_non_identifier_relationships(self):
        for rel_meta in self._reasoner.bound_relationships():
            if not self._reasoner.is_identifier_relationship(rel_meta.relationship):
                relationship = rel_meta.relationship
                if relationship.arity() >= 2:  # arity 2+ are semantic predicates
                    self._gen_semantic_predicate(rel_meta)

    def _gen_preferred_id(self, binding: Union[IdentifierBinding, SubtypeBinding]):
        # =
        # Simple case: a single role, hence the preferred id is the role itself.
        #
        # We generate the following rules:
        #
        #   def {concept}:id(v, c): ...
        # =
        ctor_role = self._reasoner.lookup_binding_role(binding)
        value_concept = ctor_role.player()
        val_role = ctor_role.sibling()
        assert val_role is not None
        entity_concept = val_role.player()

        # find the relations that have been created
        pref_id_relation = None
        identifies_relation = None
        for reading in ctor_role.part_of.readings():
            roles = reading.roles
            rel_name = reading.rel_name
            if roles[0] == ctor_role:
                if isinstance(value_concept, EntityType) or isinstance(value_concept, ValueType):
                    identifies_relation = getattr(value_concept, rel_name)
            else:
                assert isinstance(entity_concept, EntityType)
                pref_id_relation = entity_concept[rel_name]

        rel_nm = self._fqn_preferred_id(entity_concept)
        assert pref_id_relation is not None
        with pref_id_relation:
            @rule()
            def identifier(c, i):
                row = Vars(RowId)
                self._ref_value_map(binding)(row, i)
                self._ref_constructor_entity_map(binding)(row, c)
        self._register_relation(rel_nm, pref_id_relation, self._preferred_id_relations)
        self._entity_to_id_relations[entity_concept] = pref_id_relation

        # multiple bindings can exist, but we only need to transpose once
        rel_nm = self._fqn_identifies(value_concept)
        # TODO: fix this one below
        if isinstance(value_concept, AbstractValueType) and rel_nm not in self._relations:
            identifies_relation = self._model.external_relation(rel_nm, value_concept, entity_concept)
        assert identifies_relation is not None
        if identifies_relation not in self._identifier_relations:
            assert pref_id_relation is not None
            with identifies_relation:
                @rule()
                def identifies(i, c):
                    pref_id_relation(c, i)
            self._register_relation(rel_nm, identifies_relation, self._identifier_relations)
            self._concept_to_identifies_relations[value_concept] = identifies_relation

    @staticmethod
    def _fqn_preferred_id(concept: 'EntityType'):
        return f'{camel_to_snake(concept.display())}__id'

    def _ref_identifies(self, binding: 'Binding'):
        pid_ref = binding.column.references
        if pid_ref:
            concept = self._reasoner.referenced_concept(pid_ref)
        else:
            role = self._reasoner.lookup_binding_role(binding)
            concept = role.player()
        fqn = self._fqn_identifies(concept)
        return self._lookup_relation_by_fqn(fqn)

    @staticmethod
    def _fqn_identifies(concept: 'Type'):
        if isinstance(concept, EntityType):
            concept = cast(EntityType, concept)
            ref_schema_nm = concept.ref_schema_name()
        else:
            ref_schema_nm = 'identifies'
        return f'{camel_to_snake(concept.display())}__{ref_schema_nm}'

    def _transpose(self, relation: 'Relation', fqn: str):
        if fqn in self._relations:
            return self._relations[fqn]
        if relation.arity() != 2:
            raise Exception('Transposition only supported for binary relations')
        (left_concept, right_concept) = relation.signature().types()
        rel = self._model.external_relation(fqn, right_concept, left_concept)
        with rel:
            @rule()
            def transpose(left, right):
                relation(right, left)
        self._relations[fqn] = rel
        return rel

    def _gen_entity_map(self, binding: 'Binding'):
        vt_role = self._reasoner.lookup_binding_role(binding)
        et_role = vt_role.sibling()
        assert et_role is not None
        value_concept = vt_role.player()
        entity_concept = et_role.player()
        assert isinstance(entity_concept, EntityType)
        rel_nm = self._fqn_entity_map_indexed(binding, et_role)
        rel = EntityMap(Namespace.top, rel_nm, RowId, binding.column.relation(), et_role)
        role_map = self._ref_value_map(binding) \
            if isinstance(value_concept, AbstractValueType) \
            else self._ref_constructor_entity_map(binding)
        with rel:
            @rule()
            def entity_map(row, et):
                val = Vars(value_concept)
                role_map(row, val)
                if binding.column.references:
                    self._ref_identifies(binding)(val, et)
                else:
                    _ = entity_concept ^ (val, et)  # `_=` is not strictly needed, but it makes IDEs happy ;)
        self._register_relation(rel_nm, rel, self._entity_map_relations)
        self._constructor_binding_to_value_map[binding] = rel

    def _ref_constructor_entity_map(self, binding: 'Binding'):
        role = self._reasoner.lookup_binding_role(binding)
        return self._ref_entity_map(binding, role.sibling())

    def _fqn_entity_map_indexed(self, binding: 'Binding', role: Optional['Role'] = None):
        if isinstance(binding, FilteringSubtypeBinding):
            if role is None:
                bindings_list = self._reasoner.subtype_bindings_of(binding.sub_type)
                fqn = self._fqn_entity_map(binding, binding.sub_type)
                idx = bindings_list.index(binding)
            else:
                bindings_list = self._reasoner.role_bindings_of(role)
                player = role.player()
                ctor_binding = None
                # Note: this works just for one ID binding per subtype for now
                for cand_binding in bindings_list:
                    if cand_binding.column.table == binding.column.table:
                        ctor_binding = cand_binding
                        break
                assert ctor_binding is not None
                fqn = self._fqn_entity_map(binding, player)
                idx = bindings_list.index(ctor_binding)
        elif isinstance(binding, SubtypeBinding) and role is None:
            # for now, only one ref table is supported
            fqn = self._fqn_entity_map(binding, binding.sub_type)
            idx = 0
        elif role is not None:
            bindings_list = self._reasoner.role_bindings_of(role)
            fqn = self._fqn_entity_map(binding, role.player())
            idx = bindings_list.index(binding)
        else:
            raise Exception('Role can not be optional for a binding other than FilteringSubtypeBinding')
        fqn = f'{fqn}{idx}' if idx > 0 else fqn
        return fqn

    def _ref_entity_map(self, binding: 'Binding', role: 'Role') -> EntityMap:
        fqn = self._fqn_entity_map_indexed(binding, role)
        emap = self._lookup_relation_by_fqn(fqn)
        if isinstance(emap, EntityMap):
            return emap
        else:
            raise Exception(f'Relation with fully qualified name {fqn} is not an EntityMap')

    def _ref_subtype_entity_map(self, binding: 'SubtypeBinding'):
        fqn = self._fqn_entity_map(binding, binding.sub_type)
        return self._lookup_relation_by_fqn(fqn)

    @staticmethod
    def _fqn_entity_map(binding: 'Binding', concept: 'EntityType'):
        assert isinstance(concept, EntityType)
        concept_nm = camel_to_snake(concept.display())
        source_nm = binding.column.table.physical_name()
        return f'{concept_nm}__impl__{source_nm}_row_to_{concept_nm}'

    def _gen_subtype_entity_map(self, binding: 'SubtypeBinding'):
        subtype = binding.sub_type
        rel_nm = self._fqn_entity_map_indexed(binding)
        rel = EntitySubtypeMap(Namespace.top, rel_nm, RowId, binding)
        if isinstance(binding, FilteringSubtypeBinding):
            self._gen_filtering_subtype_entity_map_rule(binding, rel)
        else:
            self._gen_subtype_entity_map_rule(binding, rel)
        self._register_relation(rel_nm, rel, self._subtype_entity_map_relations)
        if subtype not in self._subtype_to_entity_map:
            self._subtype_to_entity_map[subtype] = []
        self._subtype_to_entity_map[subtype].append(rel)

    def _gen_subtype_entity_map_rule(self, binding: 'SubtypeBinding', rel: 'EntitySubtypeMap'):
        with rel:
            @rule()
            def subtype_entity_map(row, et):
                self._ref_constructor_entity_map(binding)(row, et)

    def _gen_filtering_subtype_entity_map_rule(self, binding: 'FilteringSubtypeBinding', rel: 'EntitySubtypeMap'):
        ref_ctor_emap = self._ref_constructor_entity_map(binding)
        filter_column = binding.column
        filter_value = binding.has_value
        raw_value_type = filter_column.relation().attr().type()
        if isinstance(filter_value, EntityInstanceRelation):
            filter_type = filter_value.first()
            with rel:
                @rule()
                def filtering_subtype_entity_map(row, et):
                    fv, f = Vars(filter_type, raw_value_type)
                    ref_ctor_emap(row, et)
                    filter_column(row, fv)
                    _= filter_type^(fv, f)
                    filter_value(f)
        else:
            with rel:
                @rule()
                def filtering_subtype_entity_map(row, et):
                    fv = Vars(raw_value_type)
                    ref_ctor_emap(row, et)
                    filter_column(row, fv)
                    _= fv == filter_value

    def _gen_value_map(self, binding: 'Binding'):
        role = self._reasoner.lookup_binding_role(binding)
        rel_nm = self._fqn_value_map_indexed(binding)
        rel = ValueMap(Namespace.top, rel_nm, RowId, binding.column.relation(), role)
        converter = self._value_converter(binding.column)
        with rel:
            @rule(Annotation.INLINE)
            def value_map(row, val):
                if converter:
                    orig = Vars(binding.column.type())
                    rel.attr_view()(row, orig)
                    converter(orig, val)
                else:
                    rel.attr_view()(row, val)
        self._binding_to_value_map[binding] = rel
        self._register_relation(rel_nm, rel, self._value_map_relations)

    def _fqn_value_map(self, binding: 'Binding'):
        role = self._reasoner.lookup_binding_role(binding)
        concept = role.player()
        assert isinstance(concept, AbstractValueType)
        concept_nm = camel_to_snake(concept.display())
        source_nm = binding.column.table.physical_name()
        return f'{concept_nm}__impl__{source_nm}_row_to_{concept_nm}'

    def _fqn_value_map_indexed(self, binding: 'Binding'):
        role = self._reasoner.lookup_binding_role(binding)
        role_bindings = self._reasoner.role_bindings_of(role)
        idx = role_bindings.index(binding)
        fqn = self._fqn_value_map(binding)
        fqn = f'{fqn}{idx}' if idx > 0 else fqn
        return fqn

    def _ref_value_map(self, binding: 'Binding'):
        fqn = self._fqn_value_map_indexed(binding)
        return self._lookup_relation_by_fqn(fqn)

    def _gen_entity_population(self, entity_type: 'EntityType', lookup_rel: 'AbstractRelation'):
        rel_nm = self._fqn_entity_population(entity_type)
        rel = entity_type  # in the context, an entity type can be used as the population relation
        if lookup_rel in self._preferred_id_relations:
            val_type = self._get_last_type(lookup_rel)
            with rel:
                @rule()
                def entity_population(et):
                    row = Vars(val_type)
                    lookup_rel(et, row)
        elif lookup_rel in self._subtype_entity_map_relations or isinstance(lookup_rel, CompositeEntityMap):
            with rel:
                @rule()
                def entity_population(et):
                    row = Vars(RowId)
                    lookup_rel(row, et)
        else:
            raise Exception(f'Unsupported weaving type for relation `{lookup_rel.qualified_name()}`')
        self._register_relation(rel_nm, rel, self._entity_population_relations)

    @staticmethod
    def _fqn_entity_population(entity_type: 'EntityType'):
        # entity population name is the same as the entity type name, with the first letter uppercased
        return entity_type.display()

    def _lookup_relation_by_fqn(self, fqn: str):
        if fqn in self._relations:
            return self._relations[fqn]
        else:
            raise Exception(f'Relation with fully qualified name {fqn} not found in the model')

    def _gen_semantic_predicate(self, rel_meta: 'PotentiallyBoundRelationship'):
        relationship = rel_meta.relationship
        roles = relationship.roles()
        # Each role should either be covered by a value map (AbstractValueType role) or by an entity map
        # (EntityType role), which either can be inferred if unique exists for the EntityType or must be
        # generated by the respective EntityBinding.
        role_to_role_map = {}
        for role in roles:
            role_is_bound = role in self._reasoner.role_bindings()
            player = role.player()
            # for ValueTypes we just look up ValueMap relations
            if isinstance(player, AbstractValueType):
                if not role_is_bound:
                    raise Exception(f'ValueType role {role.name()} is not bound')
                else:
                    value_maps = [self._ref_value_map(binding) for binding in self._reasoner.role_bindings()[role]]
                    if len(value_maps) == 0:
                        raise Exception(f'ValueType role {role.name()} is not correctly bound')
                    if role not in role_to_role_map:
                        role_to_role_map[role] = OrderedSet()
                    role_to_role_map[role].update(value_maps)
            elif isinstance(player, EntityType):
                entity_maps = OrderedSet()
                if not role_is_bound:
                    maps = self._lookup_inferred_entity_maps(rel_meta.table, role)
                    entity_maps.update(maps)
                else:
                    bindings = self._reasoner.role_bindings()[role]
                    for binding in bindings:
                        if isinstance(binding, RoleBinding) and binding in self._reasoner.referent_bindings():
                            binding = self._reasoner.get_ctor_binding(binding)
                        entity_map = self._ref_constructor_entity_map(binding)
                        assert entity_map is not None
                        entity_maps.add(entity_map)
                if len(entity_maps) == 0:
                    raise Exception(f'EntityType role {role.name()} is not correctly bound')
                if role not in role_to_role_map:
                    role_to_role_map[role] = OrderedSet()
                role_to_role_map[role].update(entity_maps)
            else:
                raise Exception(f'Role {role.name()} is not bound to a ValueType or EntityType')
        # if we got all roles with role maps, we can generate the rule
        items = role_to_role_map.items()
        if len(items) != relationship.arity():
            raise Exception(f'Not all roles of relationship {relationship.pprint()} are bound')
        for reading in relationship.readings():
            concept = reading.roles[0].player()
            rel_nm = reading.rel_name
            relation = concept[rel_nm]

            role_map_combinations = self._permute_role_maps(role_to_role_map, as_in=reading)
            for role_maps in role_map_combinations:
                with relation:
                    @rule()
                    def semantic_predicate(*args):
                        row = Vars(RowId)
                        for role_map, arg in zip(role_maps, args):
                            role_map(row, arg)

    def _lookup_inferred_entity_maps(self, table, role):
        player = role.player()
        unbound_error_msg = f'EntityType({player.display()}) role {role.name()} is not bound'

        if player.is_composite():
            emaps = self._composite_entity_to_entity_maps.get(player)
            if emaps is None:
                raise Exception(unbound_error_msg)
            return emaps
        ref_scheme_concept = self._reasoner.get_ref_scheme_type(player)
        if ref_scheme_concept is not None:
            if player in self._subtype_to_entity_map:
                maps = self._subtype_to_entity_map[player]
            else:
                binding = self._reasoner.lookup_ref_binding(player, table)
                if binding is None:
                    raise Exception(unbound_error_msg)
                maps = [self._ref_subtype_entity_map(binding)]
            return maps
        else:
            binding = self._reasoner.lookup_ctor_binding(player, table)
            if binding is None:
                raise Exception(unbound_error_msg)
            return [self._ref_constructor_entity_map(binding)]

    @staticmethod
    def _permute_role_maps(role_to_role_map: dict['Role', OrderedSet['RoleMap']], as_in: 'Reading'):
        try:
            role_map_combinations = product(*(role_to_role_map[role] for role in as_in.roles))
            return [list(combination) for combination in role_map_combinations]
        except KeyError as e:
            missing_role = e.args[0]
            raise Exception(f'Cannot permute role maps to match reading `{as_in.rel_name}`, role `{missing_role.name()}` not found')

    @staticmethod
    def _get_last_type(lookup_rel: 'AbstractRelation'):
        return lookup_rel.signature().types()[-1]

    def _register_relation(self, name: str, rel: Union[AbstractRelation, EntityType],
                           population: OrderedSet['AbstractRelation'] = None):
        if isinstance(rel, RoleMap) or isinstance(rel, EntitySubtypeMap) or isinstance(rel, CompositeEntityMap):
            self._model._add_relation(rel)
        self._relations[name] = rel
        if population is not None:
            population.add(rel)

    def _cfg_decimal_size(self):
        return self._config['decimal_size']

    def _cfg_decimal_scale(self):
        return self._config['decimal_scale']



import dataclasses
from abc import abstractmethod
from typing import Optional, Any, Tuple, TypeVar, Generic, Union, List

from relationalai.early_access.dsl.bindings.common import BindableAttribute, BindableTable
from relationalai.early_access.dsl.bindings.relations import AttributeView
from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.core.relations import Relation
from relationalai.early_access.dsl.core.types.standard import String, Integer, Decimal, Date, DateTime, Boolean, \
    BigInteger, Float, RowId, Hash
from relationalai.early_access.dsl.core.types.unconstrained import UnconstrainedValueType
from relationalai.early_access.dsl.core.utils import generate_stable_uuid
from relationalai.early_access.dsl.ontologies.roles import Role
from relationalai.early_access.dsl.types.entities import EntityType
from relationalai.early_access.dsl.utils import normalize

#=
# Constants.
#=

DEFAULT_DECIMAL_SIZE = 64
DEFAULT_DECIMAL_SCALE = 8

#=
# Physical metadata for a Snowflake table.
#=

@dataclasses.dataclass(frozen=True)
class ColumnRef:
    table: str
    column: str

@dataclasses.dataclass
class ForeignKey:
    name: str
    source_columns: list[ColumnRef] = dataclasses.field(default_factory=list)
    target_columns: list[ColumnRef] = dataclasses.field(default_factory=list)

    def __hash__(self):
        return hash(self.name)

@dataclasses.dataclass
class SchemaMetadata:
    name: str
    foreign_keys: list[ForeignKey] = dataclasses.field(default_factory=list)

@dataclasses.dataclass
class ColumnMetadata:
    name: str
    datatype: str
    is_nullable: bool
    numeric_precision: Optional[int] = None
    numeric_precision_radix: Optional[int] = None
    numeric_scale: Optional[int] = None

@dataclasses.dataclass
class CsvColumnMetadata:
    name: str
    datatype: UnconstrainedValueType

@dataclasses.dataclass
class TabularMetadata:
    name: str
    columns: list[ColumnMetadata] = dataclasses.field(default_factory=list)
    foreign_keys: set[ForeignKey] = dataclasses.field(default_factory=set)

@dataclasses.dataclass(frozen=True)
class Binding:
    column: 'BindableColumn'

@dataclasses.dataclass(frozen=True)
class RoleBinding(Binding):
    role: Role

    def __str__(self):
        return f'RoleBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.role.player().display()}]'

@dataclasses.dataclass(frozen=True)
class ConceptBinding(Binding):
    """
    Represents a binding between an identifier column and a specific entity type.

    This binding could either represent a constructor binding (instances of the entity type are constructed from the
    values), referent binding (instances of the entity type are being looked up by the values), or a subtype binding
    (instances are being looked up using the parent type's ref scheme, or it acts as a constructor for the subtype).

    The binding later gets classified by the reasoner into either an IdentifierBinding or a SubtypeBinding.

    Attributes:
    entity_type (EntityType): The type of the entity that the identifier is bound to (directly or indirectly).
    """
    entity_type: EntityType

    def __str__(self):
        return f'ConceptBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.entity_type.display()}]'

@dataclasses.dataclass(frozen=True)
class IdentifierBinding(Binding):
    entity_type: EntityType

    def __str__(self):
        return f'IdentifierBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.entity_type.display()}]'

@dataclasses.dataclass(frozen=True)
class SubtypeBinding(Binding):
    sub_type: EntityType

    def __str__(self):
        return f'SubtypeBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.sub_type.display()}]'

@dataclasses.dataclass(frozen=True)
class FilteringSubtypeBinding(SubtypeBinding):
    has_value: Any

    def __str__(self):
        return f'FilteringSubtypeBinding[{self.column.table.physical_name()}:{self.column.physical_name()} == {self.has_value}]'

_sf_type_mapping = {
    'varchar': String,
    'char': String,
    'text': String,
    'date': Date,
    'datetime': DateTime,
    'timestamp_ntz': DateTime,
    'boolean': Boolean,
    'float': Float,
}
def _map_rai_type(col: ColumnMetadata) -> UnconstrainedValueType:
    datatype = col.datatype.lower()
    # TODO: do better with integer type conversion
    if datatype == 'number':
        if col.numeric_scale is not None and col.numeric_scale > 0:
            return Decimal
        elif col.numeric_precision is not None and col.numeric_precision > 9:
            return BigInteger
        else:
            return Integer
    else:
        return _sf_type_mapping[datatype]

class BindableColumn(BindableAttribute):
    _table: BindableTable
    _references: Optional[ColumnRef]
    _attr_relation: AttributeView

    def __init__(self, table: BindableTable, model):
        self._table = table
        self._model = model
        self._references = None

    def __call__(self, *args):
        if self._attr_relation is None:
            raise Exception(f'Attribute view for `{self.physical_name()}` not initialized')
        return self.relation()(*args)

    @abstractmethod
    def relation(self) -> AttributeView:
        pass

    def identifies(self, entity_type: EntityType):
        binding = ConceptBinding(column=self, entity_type=entity_type)
        self._model.binding(binding)

    def filters_subtype(self, sub_type: EntityType, by_value: Any):
        binding = FilteringSubtypeBinding(column=self, sub_type=sub_type, has_value=by_value)
        self._model.binding(binding)

    def binds(self, elm: Union[Relation, Role]):
        if isinstance(elm, Relation):
            roles = elm.reading().roles
            # this binds to the last role in binary relations
            role = roles[-1]
        elif isinstance(elm, Role):
            role = elm
        else:
            raise Exception(f'Expected Relation or Role, got {type(elm)}')
        binding = RoleBinding(role=role, column=self)
        self._model.binding(binding)

    @property
    def table(self):
        return self._table

    @property
    def references(self):
        return self._references

    @references.setter
    def references(self, ref: ColumnRef):
        self._references = ref

    @abstractmethod
    def physical_name(self) -> str:
        pass

    @abstractmethod
    def type(self) -> UnconstrainedValueType:
        pass

    def ref(self) -> ColumnRef:
        return ColumnRef(self._table.table_name, self.physical_name())

    def guid(self):
        return generate_stable_uuid(f'$attr:{self._table.physical_name()}_{self.physical_name()}')


class BindableSnowflakeColumn(BindableColumn):
    _metadata: ColumnMetadata

    def __init__(self, metadata: ColumnMetadata, table: 'SnowflakeTable', model):
        super().__init__(table, model)
        self._metadata = metadata
        self._datatype = _map_rai_type(self._metadata)
        self._attr_relation = AttributeView(Namespace.top, self)

    def relation(self) -> AttributeView:
        return self._attr_relation

    @property
    def metadata(self):
        return self._metadata

    def physical_name(self) -> str:
        return self._metadata.name

    def type(self) -> UnconstrainedValueType:
        return _map_rai_type(self._metadata)

    def decimal_scale(self):
        return self._metadata.numeric_scale

    def decimal_size(self):
        precision = self._metadata.numeric_precision
        if precision is not None:
            if precision <= 2:
                return 8
            elif 2 < precision <= 4:
                return 16
            elif 4 < precision <= 9:
                return 32
            elif 9 < precision <= 18:
                return 64
            elif 18 < precision <= 38:
                return 128
            raise ValueError(f'Precision {precision} is not supported (max: 38)')
        return precision

BUILTIN_TO_REL_CONVERSION = {
    Hash: "int(128)",
    String: "string",
    Integer: "int",
    BigInteger: "int(128)",
    Float: "float",
    Decimal: "decimal(64, 4)",
    Boolean: "bool",
    Date: "date",
    DateTime: "datetime"
}
class BindableCsvColumn(BindableColumn):
    _metadata: CsvColumnMetadata
    _column_rel_type: str

    def __init__(self, metadata: CsvColumnMetadata, table: 'CsvTable', model):
        super().__init__(table, model)
        self._metadata = metadata
        root_type = metadata.datatype.root_unconstrained_type()
        self._column_rel_type = BUILTIN_TO_REL_CONVERSION[root_type] if root_type in BUILTIN_TO_REL_CONVERSION else "string"
        self._attr_relation = AttributeView(Namespace.top, self)

    def relation(self) -> AttributeView:
        return self._attr_relation

    @property
    def metadata(self):
        return self._metadata

    def physical_name(self) -> str:
        return self._metadata.name

    def type(self) -> UnconstrainedValueType:
        return self._metadata.datatype

    def rel_type(self):
        return self._column_rel_type

    def decimal_scale(self) -> int:
        return DEFAULT_DECIMAL_SCALE

    def decimal_size(self) -> int:
        return DEFAULT_DECIMAL_SIZE

T = TypeVar("T", bound=BindableColumn)

class AbstractBindableTable(BindableTable,  Generic[T]):
    _columns: dict[str, T] = dataclasses.field(default_factory=dict)
    _foreign_keys: set[ForeignKey]

    def __init__(self, name: str, foreign_keys: set[ForeignKey]):
        super().__init__(name)
        self._foreign_keys = foreign_keys

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        if key in self._columns:
            return self._columns[key]
        raise AttributeError(f'Table "{self.table_name}" has no column named "{key}"')

    def foreign_key(self, *refs: Tuple[BindableColumn, BindableColumn]):
        source_columns = []
        target_columns = []

        for source, target in refs:
            source_columns.append(ColumnRef(source.table.table_name, source.physical_name()))
            target_columns.append(ColumnRef(target.table.table_name, target.physical_name()))

        source_col_names = "_".join(col.column for col in source_columns)
        target_col_names = "_".join(col.column for col in target_columns)

        fk_name = f"fk_{source_col_names}__to__{target_col_names}"

        fk = ForeignKey(fk_name, source_columns, target_columns)
        self._foreign_keys.add(fk)
        self._process_foreign_key(fk)

    @abstractmethod
    def key_type(self) -> UnconstrainedValueType:
        pass

    @abstractmethod
    def physical_name(self) -> str:
        pass

    def columns(self):
        return self._columns

    def _process_foreign_keys(self):
        for fk in self._foreign_keys:
            self._process_foreign_key(fk)

    def _process_foreign_key(self, fk):
        # TODO : this doesn't work for composite FKs
        for col in fk.source_columns:
            target_col = fk.target_columns[0]
            self._columns[col.column].references = target_col


class SnowflakeTable(AbstractBindableTable[BindableSnowflakeColumn]):

    def __init__(self, metadata: TabularMetadata, model):
        super().__init__(metadata.name, metadata.foreign_keys)
        self._columns = {col.name: BindableSnowflakeColumn(col, self, model) for col in metadata.columns}
        self._process_foreign_keys()

    def __str__(self):
        # returns the name of the table, as well as the columns and their types
        return self.table_name + ':\n' + '\n'.join(
            [f' {col.metadata.name} {col.metadata.datatype}' for _, col in self._columns.items()]
        ) + '\n' + '\n'.join(
            [f' {fk.source_columns} -> {fk.target_columns}' for fk in self._foreign_keys]
        )

    def physical_name(self):
        # physical relation name is always in the form of `{database}_{schema}_{table}
        return self.table_name.lower().replace('.', '_')

    def key_type(self) -> UnconstrainedValueType:
        return RowId


class CsvTable(AbstractBindableTable[BindableCsvColumn]):
    _csv_data: List[str]

    def __init__(self, name: str, schema: dict[str, UnconstrainedValueType], model):
        super().__init__(name, set())
        self._model = model
        self._csv_data = list()
        self._columns = {column_name: BindableCsvColumn(CsvColumnMetadata(column_name, column_type), self, model)
                         for column_name, column_type in schema.items()}

    def __str__(self):
        # returns the name of the table, as well as the columns and their types
        return self.table_name + ':\n' + '\n'.join(
            [f' {col.metadata.name} {col.metadata.datatype.root_unconstrained_type()}' for _, col in self._columns.items()]
        ) + '\n' + '\n'.join(
            [f' {fk.source_columns} -> {fk.target_columns}' for fk in self._foreign_keys]
        )

    @property
    def csv_data(self) -> List[str]:
        return self._csv_data

    def physical_name(self) -> str:
        return self.table_name.lower()

    def key_type(self) -> UnconstrainedValueType:
        return RowId

    def data(self, csv_data: str):
        index = len(self._csv_data)
        normalized_csv_data = normalize(csv_data)
        self._csv_data.append(normalized_csv_data)
        rel = self._generate_load_data_rule(index, normalized_csv_data)
        self._model._add_rel_raw_source(f'{self.table_name}_{index}', rel)

    def _generate_load_data_rule(self, index: int, csv_data: str):
        config_name = f"{self.table_name}_{index}_config"
        schema_elements = [
            f'def {config_name}[:schema, :{column_name}]: "{column.rel_type()}"'
            for column_name, column in self._columns.items()
        ]
        config_schema = "\n".join(schema_elements)
        config_data = f'def {config_name}[:data]:\n"""{csv_data}"""'
        insert_statement = f"def insert[:{self.table_name}]: load_csv[{config_name}]"
        return f"{config_schema}\n{config_data}\n{insert_statement}"